@@ -76,6 +76,7 @@ func1(void *arg)
 	perror("func1 fopen");
 	pthread_exit("botch");
     }
+    setlinebuf(f);
 
     j = pmUseContext(ctx1);
     if ( j < 0) {
@@ -109,6 +110,7 @@ func2(void *arg)
 	perror("func2 fopen");
 	pthread_exit("botch");
     }
+    setlinebuf(f);
 
     j = pmUseContext(ctx2);
     if ( j < 0) {
@@ -162,6 +164,7 @@ func3(void *arg)
 	perror("func3 fopen");
 	pthread_exit("botch");
     }
+    setlinebuf(f);
 
     j = pmUseContext(ctx3);
     if ( j < 0) {
@@ -223,6 +226,23 @@ func3(void *arg)
     return(NULL);	/* pthread done */
 }
 
+static void
+wait_for_thread(char *name, pthread_t tid)
+{
+    int		sts;
+    char	*msg;
+
+    sts = pthread_join(tid, (void *)&msg);
+    if (sts == 0) {
+	if (msg == PTHREAD_CANCELED)
+	    printf("thread %s: pthread_join: cancelled?\n", name);
+	else if (msg != NULL)
+	    printf("thread %s: pthread_join: %s\n", name, msg);
+    }
+    else
+	printf("thread %s: pthread_join: error: %s\n", name, strerror(sts));
+}
+
 int
 main(int argc, char **argv)
 {
@@ -230,7 +250,6 @@ main(int argc, char **argv)
     pthread_t	tid2;
     pthread_t	tid3;
     int		sts;
-    char	*msg;
     int		errflag = 0;
     int		c;
     int		i;
@@ -331,12 +350,9 @@ main(int argc, char **argv)
 	exit(1);
     }
 
-    pthread_join(tid1, (void *)&msg);
-    if (msg != NULL) printf("tid1: %s\n", msg);
-    pthread_join(tid2, (void *)&msg); 
-    if (msg != NULL) printf("tid2: %s\n", msg);
-    pthread_join(tid3, (void *)&msg); 
-    if (msg != NULL) printf("tid3: %s\n", msg);
+    wait_for_thread("tid1", tid1);
+    wait_for_thread("tid2", tid2);
+    wait_for_thread("tid3", tid3);
 
     exit(0);
 }
