@@ -494,6 +494,7 @@ typedef enum {
     PM_HASH_WALK_DELETE_STOP,
 } __pmHashWalkState;
 PCP_CALL extern void __pmHashInit(__pmHashCtl *);
+PCP_CALL extern int __pmHashPreAlloc(int, __pmHashCtl *);
 typedef __pmHashWalkState(*__pmHashWalkCallback)(const __pmHashNode *, void *);
 PCP_CALL extern void __pmHashWalkCB(__pmHashWalkCallback, void *, const __pmHashCtl *);
 PCP_CALL extern __pmHashNode *__pmHashWalk(__pmHashCtl *, __pmHashWalkState);
@@ -502,6 +503,7 @@ PCP_CALL extern int __pmHashAdd(unsigned int, void *, __pmHashCtl *);
 PCP_CALL extern int __pmHashDel(unsigned int, void *, __pmHashCtl *);
 PCP_CALL extern void __pmHashClear(__pmHashCtl *);
 
+
 /*
  * Host specification allowing one or more pmproxy host, and port numbers
  * within the one string, i.e. pmcd host specifications of the form:
@@ -1238,6 +1240,10 @@ PCP_CALL extern void __pmServerCloseRequestPorts(void);
 PCP_CALL extern void __pmServerDumpRequestPorts(FILE *);
 PCP_CALL extern char *__pmServerRequestPortString(int, char *, size_t);
 
+/* service manager notifications */
+PCP_CALL extern int __pmServerNotifyServiceManagerReady(pid_t);
+PCP_CALL extern int __pmServerNotifyServiceManagerStopping(pid_t);
+
 /* Service broadcasting, for servers. */
 typedef struct __pmServerPresence __pmServerPresence;
 PCP_CALL extern __pmServerPresence *__pmServerAdvertisePresence(const char *, int);
