@@ -558,10 +558,36 @@ insert_cache(hdr_t *h, const char *name, int inst, int *sts)
     entry_t	*last_e = NULL;
     char	*dup;
     int		i;
-    int		hashlen = get_hashlen(h, name);
+    int		hashlen;
 
     *sts = 0;
 
+    /*
+     * a NULL name is not good ...
+     */
+    if (name == NULL) {
+	char	strbuf[20];
+	pmNotifyErr(LOG_ERR, 
+	     "insert_cache: indom %s: NULL instance name for inst %d\n",
+	     pmInDomStr_r(h->indom, strbuf, sizeof(strbuf)), inst);
+	*sts = PM_ERR_GENERIC;
+	return NULL;
+    }
+    /*
+     * it is OK for PMDA_CACHE_STRINGS type caches to have zero-length
+     * names (strings), but not instance domains ...
+     */
+    if ((h->hstate & CACHE_STRINGS) == 0 && strlen(name) == 0) {
+	char	strbuf[20];
+	pmNotifyErr(LOG_ERR, 
+	     "insert_cache: indom %s: zero length instance name for inst %d\n",
+	     pmInDomStr_r(h->indom, strbuf, sizeof(strbuf)), inst);
+	*sts = PM_ERR_GENERIC;
+	return NULL;
+    }
+
+    hashlen = get_hashlen(h, name);
+
     if (inst != PM_IN_NULL) {
 	/*
 	 * Check if instance id or instance name already in cache
@@ -577,9 +603,10 @@ insert_cache(hdr_t *h, const char *name, int inst, int *sts)
 	e = find_entry(h, NULL, inst, sts);
 	if (e != NULL) {
 	    if (name_eq(e, name, hashlen) != 1) {
+		char	strbuf[20];
 		/* instance id the same, different name */
 		if (pmDebugOptions.indom) {
-		    fprintf(stderr, "pmdaCache: store: indom %s: instance %d ", pmInDomStr(h->indom), e->inst);
+		    fprintf(stderr, "pmdaCache: store: indom %s: instance %d ", pmInDomStr_r(h->indom, strbuf, sizeof(strbuf)), e->inst);
 		    fprintf(stderr, " in cache, name \"%s\" does not match new entry \"%s\"\n", e->name, name);
 		}
 		*sts = PM_ERR_INST;
@@ -589,7 +616,8 @@ insert_cache(hdr_t *h, const char *name, int inst, int *sts)
 	e = find_entry(h, name, PM_IN_NULL, sts);
 	if (e != NULL) {
 	    if (pmDebugOptions.indom) {
-		fprintf(stderr, "pmdaCacheStoreKey: indom %s: instance \"%s\"", pmInDomStr(h->indom), e->name);
+		char	strbuf[20];
+		fprintf(stderr, "pmdaCacheStoreKey: indom %s: instance \"%s\"", pmInDomStr_r(h->indom, strbuf, sizeof(strbuf)), e->name);
 		fprintf(stderr, " in cache, id %d does not match new entry %d\n", e->inst, inst);
 	    }
 	    *sts = PM_ERR_INST;
@@ -819,7 +847,7 @@ load_cache(hdr_t *h)
 	    if ((key = malloc(keylen)) == NULL) {
 		pmNotifyErr(LOG_ERR, 
 		     "load_cache: indom %s: unable to allocate memory for keylen=%d",
-		     pmInDomStr(h->indom), keylen);
+		     pmInDomStr_r(h->indom, strbuf, sizeof(strbuf)), keylen);
 		fclose(fp);
 		return PM_ERR_GENERIC;
 	    }
@@ -971,7 +999,8 @@ store(pmInDom indom, int flags, const char *name, pmInDom inst, int keylen, cons
 
 	if (flags != PMDA_CACHE_ADD) {
 	    if (pmDebugOptions.indom) {
-		fprintf(stderr, "pmdaCache store: indom %s: instance \"%s\"", pmInDomStr(indom), name);
+		char	strbuf[20];
+		fprintf(stderr, "pmdaCache store: indom %s: instance \"%s\"", pmInDomStr_r(indom, strbuf, sizeof(strbuf)), name);
 		if (inst != PM_IN_NULL)
 		    fprintf(stderr, " (%d)", inst);
 		fprintf(stderr, " not in cache: flags=%d not allowed\n", flags);
@@ -999,9 +1028,10 @@ store(pmInDom indom, int flags, const char *name, pmInDom inst, int keylen, cons
 	    e->keylen = keylen;
 	    if (keylen > 0) {
 		if ((e->key = malloc(keylen)) == NULL) {
+		    char	strbuf[20];
 		    pmNotifyErr(LOG_ERR, 
 			 "store: indom %s: unable to allocate memory for keylen=%d",
-			 pmInDomStr(indom), keylen);
+			 pmInDomStr_r(indom, strbuf, sizeof(strbuf)), keylen);
 		    return PM_ERR_GENERIC;
 		}
 		memcpy(e->key, key, keylen);
@@ -1062,6 +1092,7 @@ pmdaCacheStoreKey(pmInDom indom, int flags, const char *name, int keylen, const
     entry_t	*e;
     const char	*mykey;
     int		mykeylen;
+    char	strbuf[20];
 
     if (indom == PM_INDOM_NULL)
 	return PM_ERR_INDOM;
@@ -1092,7 +1123,7 @@ pmdaCacheStoreKey(pmInDom indom, int flags, const char *name, int keylen, const
 	 */
 	if (key_eq(e, mykeylen, mykey) == 0) {
 	    if (pmDebugOptions.indom) {
-		fprintf(stderr, "pmdaCacheStoreKey: indom %s: instance \"%s\" (%d) in cache ", pmInDomStr(indom), e->name, e->inst);
+		fprintf(stderr, "pmdaCacheStoreKey: indom %s: instance \"%s\" (%d) in cache ", pmInDomStr_r(indom, strbuf, sizeof(strbuf)), e->name, e->inst);
 		KeyStr(stderr, e->keylen, (const char *)e->key);
 		fprintf(stderr, " does not match new entry ");
 		KeyStr(stderr, mykeylen, mykey);
@@ -1134,7 +1165,7 @@ pmdaCacheStoreKey(pmInDom indom, int flags, const char *name, int keylen, const
 		if (key_eq(e, mykeylen, mykey) == 1)
 		    break;
 		if (pmDebugOptions.indom) {
-		    fprintf(stderr, "pmdaCacheStoreKey: indom %s: instance \"%s\" (%d) in cache, ", pmInDomStr(indom), e->name, e->inst);
+		    fprintf(stderr, "pmdaCacheStoreKey: indom %s: instance \"%s\" (%d) in cache, ", pmInDomStr_r(indom, strbuf, sizeof(strbuf)), e->name, e->inst);
 		    KeyStr(stderr, e->keylen, (const char *)e->key);
 		    fprintf(stderr, " does not match new entry ");
 		    KeyStr(stderr, mykeylen, mykey);
@@ -1144,7 +1175,7 @@ pmdaCacheStoreKey(pmInDom indom, int flags, const char *name, int keylen, const
 	    }
 	    else if (key_eq(e, mykeylen, mykey) == 1) {
 		if (pmDebugOptions.indom) {
-		    fprintf(stderr, "pmdaCacheStoreKey: indom %s: instance %d ", pmInDomStr(indom), e->inst);
+		    fprintf(stderr, "pmdaCacheStoreKey: indom %s: instance %d ", pmInDomStr_r(indom, strbuf, sizeof(strbuf)), e->inst);
 		    KeyStr(stderr, e->keylen, (const char *)e->key);
 		    fprintf(stderr, " in cache, name \"%s\" does not match new entry \"%s\"\n", e->name, name);
 		}
@@ -1155,7 +1186,7 @@ pmdaCacheStoreKey(pmInDom indom, int flags, const char *name, int keylen, const
 	    /* failed after MAX_HASH_TRY rehash attempts ... */
 	    pmNotifyErr(LOG_ERR, 
 		 "pmdaCacheStoreKey: indom %s: unable allocate a new id for instance \"%s\" based on a key of %d bytes\n",
-		 pmInDomStr(h->indom), name, keylen);
+		 pmInDomStr_r(h->indom, strbuf, sizeof(strbuf)), name, keylen);
 	    return PM_ERR_GENERIC;
 	}
 
@@ -1388,10 +1419,11 @@ int pmdaCacheLookupKey(pmInDom indom, const char *name, int keylen, const void *
     }
 
     if (pmDebugOptions.indom) {
-	    fprintf(stderr, "pmdaCacheLookupKey: indom %s: ", pmInDomStr(h->indom));
-	    KeyStr(stderr, mykeylen, mykey);
-	    fprintf(stderr, ": no matching key in cache\n");
-	}
+	char	strbuf[20];
+	fprintf(stderr, "pmdaCacheLookupKey: indom %s: ", pmInDomStr_r(h->indom, strbuf, sizeof(strbuf)));
+	KeyStr(stderr, mykeylen, mykey);
+	fprintf(stderr, ": no matching key in cache\n");
+    }
     return PM_ERR_INST;
 }
 
