@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2017-2019 Red Hat.
+ * Copyright (c) 2017-2020 Red Hat.
  *
  * This library is free software; you can redistribute it and/or modify it
  * under the terms of the GNU Lesser General Public License as published
@@ -21,9 +21,6 @@
 #include "batons.h"
 #include "slots.h"
 #include "maps.h"
-#ifdef HAVE_REGEX_H
-#include <regex.h>
-#endif
 #include <fnmatch.h>
 
 #define SHA1SZ		20	/* internal sha1 hash buffer size in bytes */
@@ -372,25 +369,6 @@ extract_sha1(seriesQueryBaton *baton, pmSID series,
     return 0;
 }
 
-static int
-extract_time(seriesQueryBaton *baton, pmSID series,
-		redisReply *reply, sds *stamp)
-{
-    sds			msg, val;
-    char		*point;
-
-    if (reply->type == REDIS_REPLY_STRING) {
-	val = sdscpylen(*stamp, reply->str, reply->len);
-	if ((point = strchr(val, '-')) != NULL)
-	    *point = '.';
-	*stamp = val;
-	return 0;
-    }
-    infofmt(msg, "expected string timestamp in series %s", series);
-    batoninfo(baton, PMLOG_RESPONSE, msg);
-    return -EPROTO;
-}
-
 /*
  * Report a timeseries result - timestamps and (instance) values
  */
@@ -428,60 +406,199 @@ series_instance_reply(seriesQueryBaton *baton, sds series,
 }
 
 static int
-series_result_reply(seriesQueryBaton *baton, sds series, pmSeriesValue *value,
-		int nelements, redisReply **elements)
+extract_time(seriesQueryBaton *baton, pmSID series,
+		redisReply *reply, sds *stamp, pmTimespec *ts)
 {
-    redisReply		*reply;
-    sds			msg;
-    int			i, sts;
+    sds			msg, val;
+    char		*point = NULL;
+    __uint64_t		milliseconds, fractions, crossover;
 
-    /* expecting timestamp:valueset pairs, then instance:value pairs */
-    if (nelements % 2) {
-	infofmt(msg, "expected time:valueset pairs in %s XRANGE", series);
-	batoninfo(baton, PMLOG_RESPONSE, msg);
-	return -EPROTO;
-    }
-
-    for (i = 0; i < nelements; i += 2) {
-	reply = elements[i+1];
-	if ((sts = extract_time(baton, series, elements[i],
-				&value->timestamp)) < 0) {
-	    baton->error = sts;
-	} else if (reply->type != REDIS_REPLY_ARRAY) {
-	    infofmt(msg, "expected value array for series %s %s (type=%s)",
-			series, XRANGE, redis_reply_type(reply));
-	    batoninfo(baton, PMLOG_RESPONSE, msg);
-	    baton->error = -EPROTO;
-	} else if ((sts = series_instance_reply(baton, series, value,
-				reply->elements, reply->element)) < 0) {
-	    baton->error = sts;
+    if (reply->type == REDIS_REPLY_STRING) {
+	val = sdscpylen(*stamp, reply->str, reply->len);
+	milliseconds = strtoull(val, &point, 0);
+	if (point && *point == '-') {
+	    *point = '.';
+	    fractions = strtoull(point + 1, NULL, 0);
+	} else {
+	    fractions = 0;
 	}
+	crossover = milliseconds % 1000;
+	ts->tv_sec = milliseconds / 1000;
+	ts->tv_nsec = (fractions * 1000) + (crossover * 1000000);
+	*stamp = val;
+	return 0;
     }
+    infofmt(msg, "expected string timestamp in series %s", series);
+    batoninfo(baton, PMLOG_RESPONSE, msg);
+    return -EPROTO;
+}
+
+static inline int
+pmTimespec_cmp(pmTimespec *a, pmTimespec *b)
+{
+    if (a->tv_sec != b->tv_sec)
+	return (a->tv_sec > b->tv_sec) ? 1 : -1;
+    if (a->tv_nsec > b->tv_nsec)
+	return 1;
+    if (a->tv_nsec < b->tv_nsec)
+	return -1;
     return 0;
 }
 
+static inline void
+pmTimespec_add(pmTimespec *t1, pmTimespec *t2)
+{
+    __int64_t		sec = t1->tv_sec + t2->tv_sec;
+    __int64_t		nsec = t1->tv_nsec + t2->tv_nsec;
+
+    if (nsec >= 1000000000) {
+        nsec -= 1000000000;
+        sec++;
+    }
+    t1->tv_sec = sec;
+    t1->tv_nsec = nsec;
+}
+
+typedef struct seriesSampling {
+    unsigned int	setup;		/* one-pass calculation flag */
+    unsigned int	subsampling;	/* sample interval requested */
+    unsigned int	count;		/* N sampled values, so far */
+    pmSeriesValue	value;		/* current sample being built */
+    pmTimespec		goal;		/* 'ideal' sample timestamp */
+    pmTimespec		delta;		/* sampling interval (step) */
+    pmTimespec		next_timespec;	/* from the following sample */
+    sds			next_timestamp;	/* from the following sample */
+} seriesSampling;
+
+static int
+use_next_sample(seriesSampling *sampling)
+{
+    /* if the next timestamp is past our goal, use the current value */
+    if (pmTimespec_cmp(&sampling->next_timespec, &sampling->goal) > 0) {
+	/* selected a value for this interval so move the goal posts */
+	pmTimespec_add(&sampling->goal, &sampling->delta);
+	return 0;
+    }
+    return 1;
+}
+
 static void
 series_values_reply(seriesQueryBaton *baton, sds series,
-		int nelements, redisReply **elements, void *arg)
+		int nsamples, redisReply **samples, void *arg)
 {
-    pmSeriesValue	value;
-    redisReply		*reply;
-    int			i, sts;
+    seriesSampling	sampling = {0};
+    redisReply		*reply, *sample, **elements;
+    timing_t		*tp = &baton->u.query.timing;
+    int			n, sts, next, nelements;
+    sds			msg, save_timestamp;
 
-    value.timestamp = sdsempty();
-    value.series = sdsempty();
-    value.data = sdsempty();
+    sampling.value.timestamp = sdsempty();
+    sampling.value.series = sdsempty();
+    sampling.value.data = sdsempty();
 
-    for (i = 0; i < nelements; i++) {
-	reply = elements[i];
-	if ((sts = series_result_reply(baton, series, &value,
-				reply->elements, reply->element)) < 0)
+    /* iterate over the 'samples' array */
+    for (n = 0; n < nsamples; n++) {
+	sample = samples[n];
+	if ((nelements = sample->elements) == 0)
+	    continue;
+	elements = sample->element;
+
+	/* expecting timestamp:valueset pairs, then instance:value pairs */
+	if (nelements % 2) {
+	    infofmt(msg, "expected time:valueset pairs in %s XRANGE", series);
+	    batoninfo(baton, PMLOG_RESPONSE, msg);
+	    sts = -EPROTO;
+	    break;
+	}
+
+	/* verify the instance:value pairs array before proceeding */
+	reply = elements[1];
+	if (reply->type != REDIS_REPLY_ARRAY) {
+	    infofmt(msg, "expected value array for series %s %s (type=%s)",
+			series, XRANGE, redis_reply_type(reply));
+	    batoninfo(baton, PMLOG_RESPONSE, msg);
+	    baton->error = -EPROTO;
+	    break;
+	}
+
+	/* setup state variables used internally during selection process */
+	if (sampling.setup == 0 && (tp->delta.tv_sec || tp->delta.tv_usec)) {
+	    /* 'next' is a nanosecond precision time interval to step with */
+	    sampling.delta.tv_sec = tp->delta.tv_sec;
+	    sampling.delta.tv_nsec = tp->delta.tv_usec * 1000;
+
+	    /* extract the first timestamp to kickstart the comparison process */
+	    if ((sts = extract_time(baton, series, elements[0],
+					&sampling.value.timestamp,
+					&sampling.value.ts)) < 0) {
+		baton->error = sts;
+		break;
+	    }
+	    /* 'goal' is the first target interval as an absolute timestamp */
+	    if (tp->start.tv_sec || tp->start.tv_usec) {
+		sampling.goal.tv_sec = tp->start.tv_sec;
+		sampling.goal.tv_nsec = tp->start.tv_usec * 1000;
+	    } else {
+		sampling.goal = sampling.value.ts;
+	    }
+	    sampling.goal.tv_nsec++;	/* ensure we use first sample */
+
+	    sampling.next_timestamp = sdsempty();
+	    sampling.subsampling = 1;
+	}
+	sampling.setup = 1;
+
+	if (sampling.subsampling == 0) {
+	    if ((sts = extract_time(baton, series, elements[0],
+					&sampling.value.timestamp,
+					&sampling.value.ts)) < 0) {
+		baton->error = sts;
+		continue;
+	    }
+	} else if ((next = n + 1) < nsamples) {
+	    /*
+	     * Compare this point and the next to the ideal based on delta;
+	     * skip over returning this value if the next one looks better.
+	     */
+	    elements = samples[next]->element;
+	    if ((sts = extract_time(baton, series, elements[0],
+					&sampling.next_timestamp,
+					&sampling.next_timespec)) < 0) {
+		baton->error = sts;
+		continue;
+	    } else if ((sts = use_next_sample(&sampling)) == 1) {
+		goto next_sample;
+	    } else if (sts == -1) {		/* sampling reached the end */
+		goto last_sample;
+	    } /* else falls through and may call user-supplied callback */
+	}
+
+	/* check whether a user-requested sample count has been reached */
+	if (tp->count && sampling.count++ >= tp->count)
+	    break;
+
+	if ((sts = series_instance_reply(baton, series, &sampling.value,
+				reply->elements, reply->element)) < 0) {
 	    baton->error = sts;
+	    goto last_sample;
+	}
+
+	if (sampling.subsampling == 0)
+	    continue;
+next_sample:
+	/* carefully swap time strings to avoid leaking memory */
+	save_timestamp = sampling.next_timestamp;
+	sampling.next_timestamp = sampling.value.timestamp;
+	sampling.value.timestamp = save_timestamp;
+	sampling.value.ts = sampling.next_timespec;
     }
 
-    sdsfree(value.timestamp);
-    sdsfree(value.series);
-    sdsfree(value.data);
+last_sample:
+    if (sampling.setup)
+	sdsfree(sampling.next_timestamp);
+    sdsfree(sampling.value.timestamp);
+    sdsfree(sampling.value.series);
+    sdsfree(sampling.value.data);
 }
 
 /*
@@ -518,7 +635,7 @@ node_series_reply(seriesQueryBaton *baton, node_t *np, int nelements, redisReply
 	    memcpy(series, reply->str, SHA1SZ);
 	    if (pmDebugOptions.series) {
 		pmwebapi_hash_str(series, hashbuf, sizeof(hashbuf));
-		printf("    %s\n", hashbuf);
+		fprintf(stderr, "    %s\n", hashbuf);
 	    }
 	    series += SHA1SZ;
 	} else {
@@ -574,7 +691,7 @@ series_intersect(series_set_t *a, series_set_t *b)
 
     qsort(large, nlarge, SHA1SZ, series_compare);
 
-    for (i = 0, cp = saved = small; i < nsmall; i++, cp ++) {
+    for (i = 0, cp = saved = small; i < nsmall; i++, cp += SHA1SZ) {
 	if (!bsearch(cp, large, nlarge, SHA1SZ, series_compare))
 	    continue;		/* no match, continue advancing cp only */
 	if (saved != cp)
@@ -591,10 +708,10 @@ series_intersect(series_set_t *a, series_set_t *b)
     if (pmDebugOptions.series && pmDebugOptions.desperate) {
 	char		hashbuf[42];
 
-	printf("Intersect result set contains %d series:\n", total);
-	for (i = 0, cp = small; i < total; cp++, i++) {
+	fprintf(stderr, "Intersect result set contains %d series:\n", total);
+	for (i = 0, cp = small; i < total; cp += SHA1SZ, i++) {
 	    pmwebapi_hash_str(cp, hashbuf, sizeof(hashbuf));
-	    printf("    %s\n", hashbuf);
+	    fprintf(stderr, "    %s\n", hashbuf);
 	}
     }
 
@@ -647,7 +764,7 @@ series_union(series_set_t *a, series_set_t *b)
     }
 
     if (pmDebugOptions.series)
-	printf("Union of large(%d) and small(%d) series\n", nlarge, nsmall);
+	fprintf(stderr, "Union of large(%d) and small(%d) series\n", nlarge, nsmall);
 
     qsort(large, nlarge, SHA1SZ, series_compare);
 
@@ -674,10 +791,10 @@ series_union(series_set_t *a, series_set_t *b)
     if (pmDebugOptions.series && pmDebugOptions.desperate) {
 	char		hashbuf[42];
 
-	printf("Union result set contains %d series:\n", total);
+	fprintf(stderr, "Union result set contains %d series:\n", total);
 	for (i = 0, cp = large; i < total; cp += SHA1SZ, i++) {
 	    pmwebapi_hash_str(cp, hashbuf, sizeof(hashbuf));
-	    printf("    %s\n", hashbuf);
+	    fprintf(stderr, "    %s\n", hashbuf);
 	}
     }
 
@@ -717,7 +834,7 @@ string_pattern_match(node_t *np, sds pattern, char *string, int length)
 	return fnmatch(pattern, string, 0) == 0;
 
     /* use either regular expression match or negation */
-    sts = regexec((const regex_t *)np->regex, string, 0, NULL, 0);
+    sts = regexec((const regex_t *)&np->regex, string, 0, NULL, 0);
     if (np->type == N_REQ)
 	return sts == 0;
     if (np->type == N_RNE)
@@ -825,6 +942,8 @@ node_pattern_reply(seriesQueryBaton *baton, node_t *np, const char *name, int ne
 out:
     if (np->cursor > 0)	/* still more to retrieve - kick off the next batch */
 	series_pattern_match(baton, np);
+    else
+	regfree((regex_t *)&np->regex);
 
     return nelements;
 }
@@ -860,7 +979,7 @@ series_prepare_maps_pattern_reply(
     } else {
 	name = left->key + sizeof("pcp:map:") - 1;
 	if (pmDebugOptions.series)
-	    printf("%s %s\n", node_subtype(np->left), np->key);
+	    fprintf(stderr, "%s %s\n", node_subtype(np->left), np->key);
 	if (node_pattern_reply(baton, np, name, reply->elements, reply->element) < 0)
 	    baton->error = -EPROTO;
     }
@@ -986,7 +1105,7 @@ series_prepare_smembers_reply(
 	baton->error = -EPROTO;
     } else {
 	if (pmDebugOptions.series)
-	    printf("%s %s\n", node_subtype(np->left), np->key);
+	    fprintf(stderr, "%s %s\n", node_subtype(np->left), np->key);
 	sts = node_series_reply(baton, np, reply->elements, reply->element);
 	if (sts < 0)
 	    baton->error = sts;
@@ -1124,7 +1243,14 @@ series_prepare_time_reply(
     series_query_end_phase(baton);
 }
 
-#define DEFAULT_VALUE_COUNT 10
+unsigned int
+series_value_count_only(timing_t *tp)
+{
+    if (tp->window.range || tp->window.delta ||
+	tp->window.start || tp->window.end)
+	return 0;
+    return tp->count;
+}
 
 static void
 series_prepare_time(seriesQueryBaton *baton, series_set_t *result)
@@ -1132,27 +1258,31 @@ series_prepare_time(seriesQueryBaton *baton, series_set_t *result)
     timing_t		*tp = &baton->u.query.timing;
     unsigned char	*series = result->series;
     seriesGetSID	*sid;
-    char		buffer[64];
-    sds			count, start, end, key, cmd;
-    unsigned int	i;
+    char		buffer[64], revbuf[64];
+    sds			start, end, key, cmd;
+    unsigned int	i, revlen = 0, reverse = 0;
+
+    /* if only 'count' is requested, work back from most recent value */
+    if ((reverse = series_value_count_only(tp)) != 0) {
+	revlen = pmsprintf(revbuf, sizeof(revbuf), "%u", reverse);
+	start = sdsnew("+");
+    } else {
+	start = sdsnew(timeval_stream_str(&tp->start, buffer, sizeof(buffer)));
+    }
 
-    start = sdsnew(timeval_stream_str(&tp->start, buffer, sizeof(buffer)));
     if (pmDebugOptions.series)
 	fprintf(stderr, "START: %s\n", start);
 
-    if (tp->end.tv_sec)
+    if (reverse)
+	end = sdsnew("-");
+    else if (tp->end.tv_sec)
 	end = sdsnew(timeval_stream_str(&tp->end, buffer, sizeof(buffer)));
     else
 	end = sdsnew("+");	/* "+" means "no end" - to the most recent */
+
     if (pmDebugOptions.series)
 	fprintf(stderr, "END: %s\n", end);
 
-    if (tp->count == 0)
-	tp->count = DEFAULT_VALUE_COUNT;
-    count = sdscatfmt(sdsempty(), "%u", tp->count);
-    if (pmDebugOptions.series)
-	fprintf(stderr, "COUNT: %u\n", tp->count);
-
     /*
      * Query cache for the time series range (groups of instance:value
      * pairs, with an associated timestamp).
@@ -1166,18 +1296,24 @@ series_prepare_time(seriesQueryBaton *baton, series_set_t *result)
 
 	key = sdscatfmt(sdsempty(), "pcp:values:series:%S", sid->name);
 
-	/* XREAD key t1 t2 [COUNT count] */
-	cmd = redis_command(6);
-	cmd = redis_param_str(cmd, XRANGE, XRANGE_LEN);
+	/* X[REV]RANGE key t1 t2 [count N] */
+	if (reverse) {
+	    cmd = redis_command(6);
+	    cmd = redis_param_str(cmd, XREVRANGE, XREVRANGE_LEN);
+	} else {
+	    cmd = redis_command(4);
+	    cmd = redis_param_str(cmd, XRANGE, XRANGE_LEN);
+	}
 	cmd = redis_param_sds(cmd, key);
 	cmd = redis_param_sds(cmd, start);
 	cmd = redis_param_sds(cmd, end);
-	cmd = redis_param_str(cmd, "COUNT", sizeof("COUNT")-1);
-	cmd = redis_param_sds(cmd, count);
+	if (reverse) {
+	    cmd = redis_param_str(cmd, "COUNT", sizeof("COUNT")-1);
+	    cmd = redis_param_str(cmd, revbuf, revlen);
+	}
 	redisSlotsRequest(baton->slots, XRANGE, key, cmd,
 				series_prepare_time_reply, sid);
     }
-    sdsfree(count);
     sdsfree(start);
     sdsfree(end);
 }
@@ -1269,7 +1405,9 @@ series_query_report_values(void *arg)
 static int
 series_time_window(timing_t *tp)
 {
-    if (tp->window.range || tp->window.start || tp->window.end || tp->window.count)
+    if (tp->window.range ||
+	tp->window.start || tp->window.end ||
+	tp->window.count || tp->window.delta)
 	return 1;
     return 0;
 }
@@ -1687,6 +1825,107 @@ pmSeriesLabels(pmSeriesSettings *settings, int nseries, pmSID *series, void *arg
     return 0;
 }
 
+static void
+series_lookup_labelvalues_callback(
+	redisAsyncContext *c, redisReply *reply, const sds cmd, void *arg)
+{
+    seriesGetSID	*sid = (seriesGetSID *)arg;
+    seriesQueryBaton    *baton = (seriesQueryBaton *)sid->baton;
+    redisReply		*child;
+    pmSeriesLabel	label;
+    unsigned int	i;
+    sds			msg;
+    int                 sts;
+
+    seriesBatonCheckMagic(sid, MAGIC_SID, "series_lookup_labelvalues_callback");
+    seriesBatonCheckMagic(baton, MAGIC_QUERY, "series_lookup_labelvalues_callback");
+    sts = redisSlotsRedirect(baton->slots, reply, baton->info, baton->userdata,
+			    cmd, series_lookup_labelvalues_callback, arg);
+    if (sts > 0)
+	return;	/* short-circuit as command was re-submitted */
+
+    if (LIKELY(reply && reply->type == REDIS_REPLY_ARRAY)) {
+	label.name = sid->name;
+	label.value = sdsempty();
+
+	for (i = 0; i < reply->elements; i++) {
+	    child = reply->element[i];
+	    if (child->type == REDIS_REPLY_STRING) {
+		label.value = sdscpylen(label.value, child->str, child->len);
+		baton->callbacks->on_labelmap(NULL, &label, baton->userdata);
+	    } else {
+		infofmt(msg, "bad response for string map %s (%s)",
+			HVALS, redis_reply_type(child));
+		batoninfo(baton, PMLOG_RESPONSE, msg);
+		baton->error = -EINVAL;
+	    }
+	}
+	sdsfree(label.value);
+    } else {
+	if (sts < 0) {
+	    infofmt(msg, "expected array from string map %s (reply=%s)",
+		HVALS, redis_reply_type(reply));
+	    batoninfo(baton, PMLOG_RESPONSE, msg);
+	}
+	baton->error = -EPROTO;
+    }
+    freeSeriesGetSID(sid);
+
+    series_query_end_phase(baton);
+}
+
+static void
+series_lookup_labelvalues(void *arg)
+{
+    seriesQueryBaton	*baton = (seriesQueryBaton *)arg;
+    seriesGetSID	*sid;
+    sds			cmd, key;
+    unsigned char	hash[20];
+    char		buffer[42];
+    unsigned int	i;
+
+    seriesBatonCheckMagic(baton, MAGIC_QUERY, "series_lookup_labelvalues");
+    seriesBatonCheckCount(baton, "series_lookup_labelvalues");
+
+    for (i = 0; i < baton->u.lookup.nseries; i++) {
+	sid = &baton->u.lookup.series[i];
+	seriesBatonReference(baton, "series_lookup_labelvalues");
+
+	pmwebapi_string_hash(hash, sid->name, sdslen(sid->name));
+	key = sdscatfmt(sdsempty(), "pcp:map:label.%s.value",
+			    pmwebapi_hash_str(hash, buffer, sizeof(buffer)));
+	cmd = redis_command(2);
+	cmd = redis_param_str(cmd, HVALS, HVALS_LEN);
+	cmd = redis_param_sds(cmd, key);
+	redisSlotsRequest(baton->slots, HVALS, key, cmd,
+			series_lookup_labelvalues_callback, sid);
+    }
+}
+
+int
+pmSeriesLabelValues(pmSeriesSettings *settings, int nlabels, pmSID *labels, void *arg)
+{
+    seriesQueryBaton	*baton;
+    size_t		bytes;
+    unsigned int	i = 0;
+
+    if (nlabels <= 0)
+	return -EINVAL;
+    bytes = sizeof(seriesQueryBaton) + (nlabels * sizeof(seriesGetSID));
+    if ((baton = calloc(1, bytes)) == NULL)
+	return -ENOMEM;
+    initSeriesQueryBaton(baton, settings, arg);
+    initSeriesGetLookup(baton, nlabels, labels, NULL, NULL);
+
+    baton->current = &baton->phases[0];
+    baton->phases[i++].func = series_lookup_services;
+    baton->phases[i++].func = series_lookup_labelvalues;
+    baton->phases[i++].func = series_lookup_finished;
+    assert(i <= QUERY_PHASES);
+    seriesBatonPhases(baton->current, i, baton);
+    return 0;
+}
+
 static int
 extract_series_desc(seriesQueryBaton *baton, pmSID series,
 		int nelements, redisReply **elements, pmSeriesDesc *desc)
@@ -1699,11 +1938,15 @@ extract_series_desc(seriesQueryBaton *baton, pmSID series,
 	return -EPROTO;
     }
 
-    /* sanity check - were we given an invalid series identifier? */
+    /* were we given a non-metric series identifier? (e.g. an instance) */
     if (elements[0]->type == REDIS_REPLY_NIL) {
-	infofmt(msg, "no descriptor for series identifier %s", series);
-	batoninfo(baton, PMLOG_ERROR, msg);
-	return -EINVAL;
+	desc->indom = sdscpylen(desc->indom, "unknown", 7);
+	desc->pmid = sdscpylen(desc->pmid, "PM_ID_NULL", 10);
+	desc->semantics = sdscpylen(desc->semantics, "unknown", 7);
+	desc->source = sdscpylen(desc->source, "unknown", 7);
+	desc->type = sdscpylen(desc->type, "unknown", 7);
+	desc->units = sdscpylen(desc->units, "unknown", 7);
+	return 0;
     }
 
     if (extract_string(baton, series, elements[0], &desc->indom, "indom") < 0)
@@ -2267,17 +2510,19 @@ parsetime(seriesQueryBaton *baton, sds string, struct timeval *result, const cha
 }
 
 static void
-parseint(seriesQueryBaton *baton, sds string, int *count, const char *source)
+parseuint(seriesQueryBaton *baton, sds string, unsigned int *vp, const char *source)
 {
+    unsigned int	value;
+    char		*endnum;
     sds			msg;
-    int			sts;
 
-    if ((sts = atoi(string)) < 0) {
+    value = (unsigned int)strtoul(string, &endnum, 10);
+    if (*endnum != '\0') {
 	infofmt(msg, "Invalid sample %s requested - %s", source, string);
 	batoninfo(baton, PMLOG_ERROR, msg);
-	baton->error = sts;
+	baton->error = -EINVAL;
     } else {
-	*count = sts;
+	*vp = value;
     }
 }
 
@@ -2359,9 +2604,9 @@ initSeriesGetValues(seriesQueryBaton *baton, int nseries, sds *series,
 	timing->end.tv_sec = INT_MAX;
     }
     if (window->count)
-	parseint(baton, window->count, &timing->count, "count");
+	parseuint(baton, window->count, &timing->count, "count");
     if (window->offset)
-	parseint(baton, window->offset, &timing->offset, "offset");
+	parseuint(baton, window->offset, &timing->offset, "offset");
     if (window->zone)
 	parsezone(baton, window->zone, &timing->zone, "timezone");
 
