@@ -14,6 +14,7 @@
 #include <assert.h>
 #include "pmapi.h"
 #include "pmda.h"
+#include "search.h"
 #include "schema.h"
 #include "discover.h"
 #include "util.h"
@@ -21,7 +22,8 @@
 
 #define STRINGIFY(s)	#s
 #define TO_STRING(s)	STRINGIFY(s)
-#define SCHEMA_VERSION	2
+#define SERIES_VERSION	2
+#define SERVER_VERSION	5
 
 extern sds		cursorcount;
 static sds		maxstreamlen;
@@ -58,97 +60,45 @@ redisScriptsInit(void)
     }
 }
 
-static int
-testReplyError(redisReply *reply, const char *server_message)
+static void
+initRedisSlotsBaton(redisSlotsBaton *baton,
+		redisInfoCallBack info, redisDoneCallBack done,
+		void *userdata, void *events, void *arg)
 {
-    return (reply && reply->type == REDIS_REPLY_ERROR &&
-	    strcmp(reply->str, server_message) == 0);
+    initSeriesBatonMagic(baton, MAGIC_SLOTS);
+    baton->info = info;
+    baton->done = done;
+    baton->version = -1;
+    baton->userdata = userdata;
+    baton->arg = arg;
 }
 
 static void
-reportReplyError(redisInfoCallBack info, void *userdata,
-	redisReply *reply, const char *format, va_list argp)
+redis_slots_finished(void *arg)
 {
-    sds			msg = sdsnew("Error: ");
+    redisSlotsBaton	*baton = (redisSlotsBaton *)arg;
 
-    msg = sdscatvprintf(msg, format, argp);
-    if (reply && reply->type == REDIS_REPLY_ERROR)
-	msg = sdscatfmt(msg, "\nRedis: %s\n", reply->str);
-    else
-	msg = sdscat(msg, "\n");
-    info(PMLOG_RESPONSE, msg, userdata);
-    sdsfree(msg);
+    seriesBatonCheckMagic(baton, MAGIC_SLOTS, "redis_slots_finished");
+
+    if (baton->error == 0)
+	baton->done(baton->arg);
+    memset(baton, 0, sizeof(*baton));
+    free(baton);
 }
 
-static int
-checkStatusReplyOK(redisInfoCallBack info, void *userdata,
-		redisReply *reply, const char *format, ...)
+void
+redis_slots_end_phase(void *arg)
 {
-    va_list		argp;
+    redisSlotsBaton	*baton = (redisSlotsBaton *)arg;
 
-    if (reply && reply->type == REDIS_REPLY_STATUS &&
-	(strcmp("OK", reply->str) == 0 || strcmp("QUEUED", reply->str) == 0))
-	return 0;
-    va_start(argp, format);
-    reportReplyError(info, userdata, reply, format, argp);
-    va_end(argp);
-    return -1;
-}
+    seriesBatonCheckMagic(baton, MAGIC_SLOTS, "redis_slots_end_phase");
 
-static int
-checkStreamReplyString(redisInfoCallBack info, void *userdata,
-	redisReply *reply, sds s, const char *format, ...)
-{
-    va_list		argp;
-
-    if (reply && reply->type == REDIS_REPLY_STRING && strcmp(s, reply->str) == 0)
-	return 0;
-    va_start(argp, format);
-    reportReplyError(info, userdata, reply, format, argp);
-    va_end(argp);
-    return -1;
-}
-
-static int
-checkArrayReply(redisInfoCallBack info, void *userdata,
-	redisReply *reply, const char *format, ...)
-{
-    va_list		argp;
-
-    if (reply && reply->type == REDIS_REPLY_ARRAY)
-	return 0;
-    va_start(argp, format);
-    reportReplyError(info, userdata, reply, format, argp);
-    va_end(argp);
-    return -1;
-}
-
-static long long
-checkIntegerReply(redisInfoCallBack info, void *userdata,
-	redisReply *reply, const char *format, ...)
-{
-    va_list		argp;
-
-    if (reply && reply->type == REDIS_REPLY_INTEGER)
-	return reply->integer;
-    va_start(argp, format);
-    reportReplyError(info, userdata, reply, format, argp);
-    va_end(argp);
-    return -1;
-}
-
-static sds
-checkStringReply(redisInfoCallBack info, void *userdata,
-	redisReply *reply, const char *format, ...)
-{
-    va_list		argp;
-
-    if (reply && reply->type == REDIS_REPLY_STRING)
-	return sdsnew(reply->str);
-    va_start(argp, format);
-    reportReplyError(info, userdata, reply, format, argp);
-    va_end(argp);
-    return NULL;
+    if (baton->error == 0) {
+	seriesPassBaton(&baton->current, baton, "redis_slots_end_phase");
+    } else {	/* fail after waiting on outstanding I/O */
+	if (seriesBatonDereference(baton, "redis_slots_end_phase"))
+	    redis_slots_finished(baton);
+    }
 }
 
 static void
@@ -937,7 +887,7 @@ redis_series_metadata(context_t *context, metric_t *metric, void *arg)
     redisSlots			*slots = baton->slots;
     instance_t			*instance;
     value_t			*value;
-    const char			*units, *indom, *pmid, *sem, *type;
+    const char			*units, *indom = NULL, *pmid, *sem, *type;
     char			ibuf[32], pbuf[32], sbuf[20], tbuf[20], ubuf[60];
     char			hashbuf[42];
     sds				cmd, key;
@@ -992,6 +942,11 @@ redis_series_metadata(context_t *context, metric_t *metric, void *arg)
 	cmd = redis_param_str(cmd, "units", sizeof("units")-1);
 	cmd = redis_param_str(cmd, units, strlen(units));
 	redisSlotsRequest(slots, HMSET, key, cmd, redis_desc_series_callback, arg);
+
+	if ((baton->flags & PM_SERIES_FLAG_TEXT) && slots->search)
+	    redis_search_text_add(slots, PM_SEARCH_TYPE_METRIC,
+				metric->names[i].sds, indom,
+				metric->oneline, metric->helptext, baton);
     }
 
     seriesBatonReference(baton, "redis_series_metadata");
@@ -1006,6 +961,14 @@ redis_series_metadata(context_t *context, metric_t *metric, void *arg)
     redisSlotsRequest(slots, SADD, key, cmd, redis_series_source_callback, arg);
 
 check_instances:
+    if (metric->desc.indom != PM_INDOM_NULL &&
+        (baton->flags & PM_SERIES_FLAG_TEXT) && slots->search) {
+	if (indom == NULL)
+	    indom = pmwebapi_indom_str(metric, ibuf, sizeof(ibuf));
+	redis_search_text_add(slots, PM_SEARCH_TYPE_INDOM, indom, indom,
+			metric->indom->oneline, metric->indom->helptext, baton);
+    }
+
     if (metric->desc.indom == PM_INDOM_NULL || metric->u.vlist == NULL) {
 	if (metric->cached == 0) {
 	    redis_series_labelset(slots, metric, NULL, baton);
@@ -1019,6 +982,13 @@ check_instances:
 	    if (instance->cached == 0 || metric->cached == 0) {
 		redis_series_instance(slots, metric, instance, baton);
 		redis_series_labelset(slots, metric, instance, baton);
+
+		if ((baton->flags & PM_SERIES_FLAG_TEXT) && slots->search) {
+		    if (indom == NULL)
+			indom = pmwebapi_indom_str(metric, ibuf, sizeof(ibuf));
+		    redis_search_text_add(slots, PM_SEARCH_TYPE_INST,
+				instance->name.sds, indom, NULL, NULL, baton);
+		}
 	    }
 	    instance->cached = 1;
 	}
@@ -1264,6 +1234,7 @@ redis_update_version_callback(
     int			sts;
 
     seriesBatonCheckMagic(baton, MAGIC_SLOTS, "redis_update_version_callback");
+
     sts = redisSlotsRedirect(baton->slots, reply, baton->info, baton->userdata,
 			     cmd, redis_update_version_callback, arg);
     if (sts > 0)
@@ -1271,14 +1242,16 @@ redis_update_version_callback(
     if (sts == 0)
 	checkStatusReplyOK(baton->info, baton->userdata, reply,
 			"%s setup", "pcp:version:schema");
-    doneRedisSlotsBaton(baton);
+    redis_slots_end_phase(baton);
 }
 
 static void
 redis_update_version(redisSlotsBaton *baton)
 {
     sds			cmd, key;
-    const char		ver[] = TO_STRING(SCHEMA_VERSION);
+    const char		ver[] = TO_STRING(SERIES_VERSION);
+
+    seriesBatonReference(baton, "redis_update_version");
 
     key = sdsnew("pcp:version:schema");
     cmd = redis_command(3);
@@ -1289,7 +1262,7 @@ redis_update_version(redisSlotsBaton *baton)
 }
 
 static void
-redis_load_version_callback(
+redis_load_series_version_callback(
 	redisAsyncContext *c, redisReply *reply, const sds cmd, void *arg)
 {
     redisSlotsBaton	*baton = (redisSlotsBaton *)arg;
@@ -1297,9 +1270,9 @@ redis_load_version_callback(
     int			sts;
     sds			msg;
 
-    seriesBatonCheckMagic(baton, MAGIC_SLOTS, "redis_load_version_callback");
+    seriesBatonCheckMagic(baton, MAGIC_SLOTS, "redis_load_series_version_callback");
     sts = redisSlotsRedirect(baton->slots, reply, baton->info, baton->userdata,
-				cmd, redis_load_version_callback, arg);
+				cmd, redis_load_series_version_callback, arg);
     if (sts > 0)
 	return;	/* short-circuit as command was re-submitted */
 
@@ -1307,11 +1280,11 @@ redis_load_version_callback(
 	baton->version = 0;	/* NIL - no version key yet */
     } else if (reply->type == REDIS_REPLY_STRING) {
 	version = (unsigned int)atoi(reply->str);
-	if (version == 0 || version == SCHEMA_VERSION) {
+	if (version == 0 || version == SERIES_VERSION) {
 	    baton->version = version;
 	} else {
-	    infofmt(msg, "unsupported schema (got v%u, expected v%u)",
-			version, SCHEMA_VERSION);
+	    infofmt(msg, "unsupported series schema (got v%u, expected v%u)",
+			version, SERIES_VERSION);
 	    batoninfo(baton, PMLOG_ERROR, msg);
 	}
     } else if (reply->type == REDIS_REPLY_ERROR) {
@@ -1328,22 +1301,96 @@ redis_load_version_callback(
     }
 
     /* set the version when none found (first time through) */
-    if (version != SCHEMA_VERSION && baton->version != -1)
+    if (version != SERIES_VERSION && baton->version != -1) {
+	/* drop reference from schema version request */
+	seriesBatonDereference(baton, "redis_load_series_version_callback");
 	redis_update_version(arg);
-    else
-	doneRedisSlotsBaton(baton);
+    } else {
+	redis_slots_end_phase(baton);
+    }
 }
 
 static void
-redis_load_version(redisSlotsBaton *baton)
+redis_load_series_version(void *arg)
 {
+    redisSlotsBaton	*baton = (redisSlotsBaton *)arg;
     sds			cmd, key;
 
+    seriesBatonReference(baton, "redis_load_series_version");
+
     key = sdsnew("pcp:version:schema");
     cmd = redis_command(2);
     cmd = redis_param_str(cmd, GETS, GETS_LEN);
     cmd = redis_param_sds(cmd, key);
-    redisSlotsRequest(baton->slots, GETS, key, cmd, redis_load_version_callback, baton);
+    redisSlotsRequest(baton->slots, GETS, key, cmd, redis_load_series_version_callback, baton);
+}
+
+static void
+redis_load_version_callback(
+	redisAsyncContext *c, redisReply *reply, const sds cmd, void *arg)
+{
+    redisSlotsBaton	*baton = (redisSlotsBaton *)arg;
+    unsigned int	server_version = 0;
+    size_t		l;
+    char		*endnum;
+    int			sts;
+    sds			msg;
+
+    seriesBatonCheckMagic(baton, MAGIC_SLOTS, "redis_load_version_callback");
+    sts = redisSlotsRedirect(baton->slots, reply, baton->info, baton->userdata,
+				cmd, redis_load_version_callback, arg);
+    if (sts > 0)
+	return;
+	
+    if (!reply) {
+	/* This situation should not happen, since we can always get server info from redis */
+	infofmt(msg, "no redis version reply");
+	batoninfo(baton, PMLOG_ERROR, msg);
+    } else if (reply->type == REDIS_REPLY_STRING) {
+	for (l = 0; l < reply->len; l++) {
+	    if (strncmp("redis_version:", reply->str+l, sizeof("redis_version:")-1) == 0) {
+		l += sizeof("redis_version:")-1;
+	    	server_version = (unsigned int)strtoul(reply->str+l, &endnum, 10);
+	    	if (*endnum != '.') {
+		    infofmt(msg, "redis server version parse error");
+		    batoninfo(baton, PMLOG_ERROR, msg);
+	    	} else if (server_version < SERVER_VERSION) {
+		    infofmt(msg, "unsupported redis server (got v%u, expected v%u or above)", 
+				server_version, SERVER_VERSION);
+	    	    batoninfo(baton, PMLOG_ERROR, msg);
+	    	}
+	    	break;
+	    }
+	    /* move to the end of this line within the reply string */
+	    while (++l < reply->len) {
+		if (reply->str[l] == '\n' || reply->str[l] == '\0')
+		    break;
+	    }
+	}
+    } else if (reply->type == REDIS_REPLY_ERROR) {
+	if (sts < 0) {
+	    infofmt(msg, "redis server version check error: %s", reply->str);
+	    batoninfo(baton, PMLOG_REQUEST, msg);
+	}
+    } else {
+	infofmt(msg, "unexpected redis server version reply type (%s)", redis_reply_type(reply));
+	batoninfo(baton, PMLOG_ERROR, msg);
+    }
+    redis_slots_end_phase(baton);
+}
+
+static void
+redis_load_version(void *arg)
+{
+    redisSlotsBaton	*baton = (redisSlotsBaton *)arg;
+    sds			cmd;
+
+    seriesBatonReference(baton, "redis_load_version");
+
+    cmd = redis_command(2);
+    cmd = redis_param_str(cmd, INFO, INFO_LEN);
+    cmd = redis_param_str(cmd, "SERVER", sizeof("SERVER")-1);
+    redisSlotsRequest(baton->slots, INFO, NULL, cmd, redis_load_version_callback, baton);
 }
 
 static int
@@ -1404,6 +1451,7 @@ redis_load_keymap_callback(
     int			i;
 
     seriesBatonCheckMagic(baton, MAGIC_SLOTS, "redis_load_keymap_callback");
+
     /* no cluster redirection checking is needed for this callback */
     sdsfree(cmd);
 
@@ -1422,19 +1470,17 @@ redis_load_keymap_callback(
 		redis_reply_type(reply));
 	batoninfo(baton, PMLOG_ERROR, msg);
     }
-
-    /* Verify pmseries schema version if previously requested */
-    if (baton->flags & SLOTS_VERSION)
-	redis_load_version(baton);
-    else
-	doneRedisSlotsBaton(baton);
+    redis_slots_end_phase(baton);
 }
 
 static void
-redis_load_keymap(redisSlotsBaton *baton)
+redis_load_keymap(void *arg)
 {
+    redisSlotsBaton	*baton = (redisSlotsBaton *)arg;
     sds			cmd;
 
+    seriesBatonReference(baton, "redis_load_keymap");
+
     cmd = redis_command(1);
     cmd = redis_param_str(cmd, COMMAND, COMMAND_LEN);
     redisSlotsRequest(baton->slots, GETS, NULL, cmd, redis_load_keymap_callback, baton);
@@ -1569,51 +1615,23 @@ redis_load_slots_callback(
 	    decodeRedisSlots(baton, reply);
 	}
     }
-
-    if (baton->flags & SLOTS_KEYMAP)
-	/* Prepare mapping of commands to key positions if needed */
-	redis_load_keymap(baton);
-    else if (baton->flags & SLOTS_VERSION)
-	/* Verify pmseries schema version if previously requested */
-	redis_load_version(baton);
-    else
-	doneRedisSlotsBaton(baton);
+    redis_slots_end_phase(baton);
 }
 
 static void
-redis_load_slots(redisSlotsBaton *baton)
+redis_load_slots(void *arg)
 {
+    redisSlotsBaton	*baton = (redisSlotsBaton *)arg;
     sds			cmd;
 
+    seriesBatonReference(baton, "redis_load_slots");
+
     cmd = redis_command(2);
     cmd = redis_param_str(cmd, CLUSTER, CLUSTER_LEN);
     cmd = redis_param_str(cmd, "SLOTS", sizeof("SLOTS")-1);
     redisSlotsRequest(baton->slots, CLUSTER, NULL, cmd, redis_load_slots_callback, baton);
 }
 
-void
-initRedisSlotsBaton(redisSlotsBaton *baton, redisSlotsFlags flags,
-		redisInfoCallBack info, redisDoneCallBack done,
-		void *userdata, void *events, void *arg)
-{
-    initSeriesBatonMagic(baton, MAGIC_SLOTS);
-    baton->info = info;
-    baton->done = done;
-    baton->flags = flags;
-    baton->version = -1;
-    baton->userdata = userdata;
-    baton->arg = arg;
-}
-
-void
-doneRedisSlotsBaton(redisSlotsBaton *baton)
-{
-    seriesBatonCheckMagic(baton, MAGIC_SLOTS, "doneRedisSlotsBaton");
-    baton->done(baton->arg);
-    memset(baton, 0, sizeof(*baton));
-    free(baton);
-}
-
 redisSlots *
 redisSlotsConnect(dict *config, redisSlotsFlags flags,
 		redisInfoCallBack info, redisDoneCallBack done,
@@ -1622,17 +1640,32 @@ redisSlotsConnect(dict *config, redisSlotsFlags flags,
     redisSlotsBaton		*baton;
     redisSlots			*slots;
     sds				msg;
+    unsigned int		i = 0;
 
     if ((baton = (redisSlotsBaton *)calloc(1, sizeof(redisSlotsBaton))) != NULL) {
 	if ((slots = redisSlotsInit(config, events)) != NULL) {
-	    initRedisSlotsBaton(baton, flags, info, done, userdata, events, arg);
+	    initRedisSlotsBaton(baton, info, done, userdata, events, arg);
 	    baton->slots = slots;
-	    redis_load_slots(baton);
+	    baton->current = &baton->phases[0];
+	    baton->phases[i++].func = redis_load_slots;
+
+	    /* Prepare mapping of commands to key positions if needed */
+	    if (flags & SLOTS_KEYMAP)
+		baton->phases[i++].func = redis_load_keymap;
+	    /* Verify pmseries schema version and create it if needed */
+	    if (flags & SLOTS_VERSION) {
+		baton->phases[i++].func = redis_load_version; /* Redis v5 */
+		baton->phases[i++].func = redis_load_series_version;
+	    }
+	    /* Register the pmsearch schema with RediSearch if needed */
+	    if (flags & SLOTS_SEARCH)
+		baton->phases[i++].func = redis_load_search_schema;
+	    baton->phases[i++].func = redis_slots_finished;
+	    assert(i <= SLOTS_PHASES);
+	    seriesBatonPhases(baton->current, i, baton);
 	    return slots;
 	}
-	doneRedisSlotsBaton(baton);
-    } else {
-	done(arg);
+	redis_slots_end_phase(baton);
     }
     infofmt(msg, "Failed to allocate memory for Redis slots");
     info(PMLOG_ERROR, msg, arg);
@@ -1706,7 +1739,7 @@ pmSeriesSetMetricRegistry(pmSeriesModule *module, mmv_registry_t *registry)
 }
 
 static void
-redisGlobalsInit(struct dict *config)
+redisSeriesInit(struct dict *config)
 {
     sds		option;
 
@@ -1732,18 +1765,26 @@ redisGlobalsInit(struct dict *config)
     }
 }
 
+void
+redisGlobalsInit(struct dict *config)
+{
+    redisSeriesInit(config);
+    redisSearchInit(config);
+    redisScriptsInit();
+    redisMapsInit();
+}
+
 int
 pmSeriesSetup(pmSeriesModule *module, void *arg)
 {
     seriesModuleData	*data = getSeriesModuleData(module);
+    redisSlotsFlags	flags;
 
     if (data == NULL)
 	return -ENOMEM;
 
     /* create global EVAL hashes and string map caches */
     redisGlobalsInit(data->config);
-    redisScriptsInit();
-    redisMapsInit();
 
     /* fast path for when Redis has been setup already */
     if (data->slots) {
@@ -1751,8 +1792,9 @@ pmSeriesSetup(pmSeriesModule *module, void *arg)
 	data->shareslots = 1;
     } else {
 	/* establish an initial connection to Redis instance(s) */
+	flags = SLOTS_VERSION | SLOTS_SEARCH;
 	data->slots = redisSlotsConnect(
-			data->config, SLOTS_VERSION, module->on_info,
+			data->config, flags, module->on_info,
 			module->on_setup, arg, data->events, arg);
 	data->shareslots = 0;
     }
@@ -1884,7 +1926,8 @@ pmDiscoverSetup(pmDiscoverModule *module, pmDiscoverCallBacks *cbs, void *arg)
     }
 
     /* create global EVAL hashes and string map caches */
-    redisGlobalsInit(data->config);
+    redisSearchInit(data->config);
+    redisSeriesInit(data->config);
     redisScriptsInit();
     redisMapsInit();
 
@@ -1893,8 +1936,10 @@ pmDiscoverSetup(pmDiscoverModule *module, pmDiscoverCallBacks *cbs, void *arg)
 
     if (access(logdir, F_OK) == 0) {
 	sts = pmDiscoverRegister(logdir, module, cbs, arg);
-	if ((data->handle = sts) >= 0)
+	if (sts >= 0) {
+	    data->handle = sts;
 	    return 0;
+	}
     }
     return -ESRCH;
 }
