@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2019 Red Hat.
+ * Copyright (c) 2019-2020 Red Hat.
  *
  * This library is free software; you can redistribute it and/or modify it
  * under the terms of the GNU Lesser General Public License as published
@@ -55,18 +55,29 @@ webgroups_lookup(pmWebGroupModule *module)
     return (struct webgroups *)module->privdata;
 }
 
+static void
+webgroup_release_context(uv_handle_t *handle)
+{
+    struct context	*context = (struct context *)handle->data;
+
+    if (pmDebugOptions.http)
+	fprintf(stderr, "releasing context %p\n", context);
+
+    pmwebapi_free_context(context);
+}
+
 static void
 webgroup_destroy_context(struct context *context, struct webgroups *groups)
 {
     context->garbage = 1;
 
     if (pmDebugOptions.http)
-	fprintf(stderr, "freeing context %p\n", context);
+	fprintf(stderr, "destroying context %p\n", context);
 
     uv_timer_stop(&context->timer);
     if (groups)
 	dictUnlink(groups->contexts, &context->randomid);
-    pmwebapi_free_context(context);
+    uv_close((uv_handle_t *)&context->timer, webgroup_release_context);
 }
 
 static void
@@ -168,10 +179,6 @@ webgroup_new_context(pmWebGroupSettings *sp, dict *params,
     cp->context = -1;
     cp->timeout = polltime;
 
-    handle = (uv_handle_t *)&cp->timer;
-    handle->data = (void *)cp;
-    uv_timer_init(groups->events, &cp->timer);
-
     if ((cp->randomid = random()) < 0 ||
 	dictFind(groups->contexts, &cp->randomid) != NULL) {
 	infofmt(*message, "random number failure on new web context");
@@ -208,6 +215,12 @@ webgroup_new_context(pmWebGroupSettings *sp, dict *params,
 	return NULL;
     }
     dictAdd(groups->contexts, &cp->randomid, cp);
+
+    /* leave until the end because uv_timer_init makes this visible in uv_run */
+    handle = (uv_handle_t *)&cp->timer;
+    handle->data = (void *)cp;
+    uv_timer_init(groups->events, &cp->timer);
+
     cp->privdata = groups;
     cp->setup = 1;
     return cp;
@@ -1296,6 +1309,7 @@ typedef struct weblookup {
     struct context	*context;
     pmWebMetric		metric;	/* keep buffers to reduce memory allocations */
     int			status;
+    sds			message;
     void		*arg;
 } weblookup_t;
 
@@ -1310,6 +1324,9 @@ webmetric_lookup(const char *name, void *arg)
     context_t		*cp = lookup->context;
     metric_t		*mp;
 
+    if (webgroup_use_context(cp, &lookup->status, &lookup->message, arg) == NULL)
+	return;
+
     /* make sure we use the original caller supplied arg now */
     arg = lookup->arg;
 
@@ -1405,6 +1422,7 @@ pmWebGroupMetric(pmWebGroupSettings *settings, sds id, dict *params, void *arg)
     if (prefix == NULL || *prefix == '\0') {
 	sts = pmTraversePMNS_r("", webmetric_lookup, &lookup);
 	if (sts >= 0) {
+	    msg = lookup.message;
 	    sts = (lookup.status < 0) ? lookup.status : 0;
 	} else {
 	    if (sts == PM_ERR_IPC)
@@ -1416,7 +1434,9 @@ pmWebGroupMetric(pmWebGroupSettings *settings, sds id, dict *params, void *arg)
     for (i = 0; i < numnames; i++) {
 	sts = pmTraversePMNS_r(names[i], webmetric_lookup, &lookup);
 	if (sts >= 0) {
-	    sts = (lookup.status < 0) ? lookup.status : 0;
+	    msg = lookup.message;
+	    if ((sts = (lookup.status < 0) ? lookup.status : 0) < 0)
+		break;
 	} else {
 	    if (sts == PM_ERR_IPC)
 		cp->setup = 0;
@@ -1568,7 +1588,9 @@ webgroup_scrape(pmWebGroupSettings *settings, context_t *cp,
 		    if (metric->labels == NULL)
 			pmwebapi_metric_hash(metric);
 		    scrape_metric_labelsets(metric, &labels);
-		    settings->callbacks.on_scrape_labels(cp->origin, &labels, arg);
+		    if (settings->callbacks.on_scrape_labels)
+			settings->callbacks.on_scrape_labels(
+					cp->origin, &labels, arg);
 		    scrape.metric.labels = labels.buffer;
 
 		    settings->callbacks.on_scrape(cp->origin, &scrape, arg);
@@ -1592,7 +1614,9 @@ webgroup_scrape(pmWebGroupSettings *settings, context_t *cp,
 		    if (instance->labels == NULL)
 			pmwebapi_instance_hash(indom, instance);
 		    scrape_instance_labelsets(metric, indom, instance, &labels);
-		    settings->callbacks.on_scrape_labels(cp->origin, &labels, arg);
+		    if (settings->callbacks.on_scrape_labels)
+			settings->callbacks.on_scrape_labels(
+					cp->origin, &labels, arg);
 		    scrape.instance.labels = labels.buffer;
 
 		    settings->callbacks.on_scrape(cp->origin, &scrape, arg);
