@@ -7,7 +7,7 @@
 ** This source-file contains functions to calculate the differences for
 ** the system-level and process-level counters since the previous sample.
 **
-** Copyright (C) 2015,2017 Red Hat.
+** Copyright (C) 2015,2017,2019-2020 Red Hat.
 ** Copyright (C) 2000-2010 Gerlof Langeveld
 **
 ** This program is free software; you can redistribute it and/or modify it
@@ -37,8 +37,8 @@ static void calcdiff(struct tstat *, struct tstat *, struct tstat *,
 ** calculate the process-activity during the last sample
 */
 void
-deviattask(struct tstat    *curtpres, int ntaskpres,
-           struct tstat    *curpexit, int nprocexit,
+deviattask(struct tstat    *curtpres, unsigned long ntaskpres,
+           struct tstat    *curpexit, unsigned long nprocexit,
 	   struct devtstat *devtstat,
 	   struct sstat    *devsstat)
 {
@@ -49,6 +49,8 @@ deviattask(struct tstat    *curtpres, int ntaskpres,
 	count_t			totusedcpu;
 	char			hashtype = 'p';
 
+	memset(&prestat, 0, sizeof prestat);
+
 	/*
 	** needed for sanity check later on...
 	*/
@@ -82,9 +84,9 @@ deviattask(struct tstat    *curtpres, int ntaskpres,
 	** create list for the sample deviations of all tasks
 	*/
  	devtstat->ntaskall = ntaskpres + nprocexit;
-	devtstat->taskall  = malloc(devtstat->ntaskall * sizeof(struct tstat));
+	devtstat->taskall  = calloc(devtstat->ntaskall, sizeof(struct tstat));
 
-	ptrverify(devtstat->taskall, "Malloc failed for %d deviated tasks\n",
+	ptrverify(devtstat->taskall, "Malloc failed for %lu deviated tasks\n",
                                   devtstat->ntaskall);
 
 	/*
@@ -328,6 +330,33 @@ deviattask(struct tstat    *curtpres, int ntaskpres,
 			netatop_exitfind(val, devstat, &prestat);
 		}
 
+		/*
+		** handle the gpu counters
+		*/
+		if (curstat->gpu.state || prestat.gpu.state) // GPU use?
+		{
+			if (curstat->gpu.state)
+				devstat->gpu.state = curstat->gpu.state;
+			else
+				devstat->gpu.state = prestat.gpu.state;
+
+			devstat->gpu.nrgpus	= curstat->gpu.nrgpus;
+			devstat->gpu.gpulist	= curstat->gpu.gpulist;
+			devstat->gpu.gpubusy 	= curstat->gpu.gpubusy;
+			devstat->gpu.membusy	= curstat->gpu.membusy;
+			devstat->gpu.timems	= curstat->gpu.timems;
+
+			devstat->gpu.memnow	= curstat->gpu.memnow;
+			devstat->gpu.memcum	= curstat->gpu.memcum -
+						  prestat.gpu.memcum;
+			devstat->gpu.sample	= curstat->gpu.sample -
+						  prestat.gpu.sample;
+		}
+		else
+		{
+			devstat->gpu.state = '\0';
+		}
+
 		if (prestat.gen.pid > 0)
 			pdb_deltask(prestat.gen.pid, prestat.gen.isproc);
 
@@ -342,9 +371,9 @@ deviattask(struct tstat    *curtpres, int ntaskpres,
 	/*
 	** create and fill other pointer lists
 	*/
-	devtstat->procall    = malloc(devtstat->nprocall *
+	devtstat->procall    = calloc(devtstat->nprocall,
 						sizeof(struct tstat *));
-	devtstat->procactive = malloc(devtstat->nprocactive *
+	devtstat->procactive = calloc(devtstat->nprocactive,
 						sizeof(struct tstat *));
 
 	ptrverify(devtstat->procall, "Malloc failed for %d processes\n",
@@ -405,6 +434,25 @@ calcdiff(struct tstat *devstat, struct tstat *curstat, struct tstat *prestat,
 	devstat->mem.vlibs    = curstat->mem.vlibs;
 	devstat->mem.vswap    = curstat->mem.vswap;
 
+	if (curstat->gpu.state || prestat->gpu.state) // GPU use?
+	{
+		if (curstat->gpu.state)
+			devstat->gpu.state = curstat->gpu.state;
+		else
+			devstat->gpu.state = prestat->gpu.state;
+
+		devstat->gpu.nrgpus  = curstat->gpu.nrgpus;
+		devstat->gpu.gpulist = curstat->gpu.gpulist;
+		devstat->gpu.gpubusy = curstat->gpu.gpubusy;
+		devstat->gpu.membusy = curstat->gpu.membusy;
+		devstat->gpu.memnow  = curstat->gpu.memnow;
+		devstat->gpu.timems  = curstat->gpu.timems;
+	}
+	else
+	{
+		devstat->gpu.state = '\0';
+	}
+
 	/*
  	** for inactive tasks, only the static values had to be copied, while
 	** all use counters have been set to zero
@@ -502,6 +550,13 @@ calcdiff(struct tstat *devstat, struct tstat *curstat, struct tstat *prestat,
 			subcount(curstat->net.udprsz, prestat->net.udprsz);
 	else
 		devstat->net.udprsz = curstat->net.udprsz;
+
+
+	if (curstat->gpu.state)
+	{
+		devstat->gpu.memcum = curstat->gpu.memcum - prestat->gpu.memcum;
+		devstat->gpu.sample = curstat->gpu.sample - prestat->gpu.sample;
+	}
 }
 
 /*
@@ -545,6 +600,9 @@ deviatsyst(struct sstat *cur, struct sstat *pre, struct sstat *dev, double inter
 	dev->cpu.all.steal = subcount(cur->cpu.all.steal, pre->cpu.all.steal);
 	dev->cpu.all.guest = subcount(cur->cpu.all.guest, pre->cpu.all.guest);
 
+	dev->cpu.all.instr = subcount(cur->cpu.all.instr, pre->cpu.all.instr);
+	dev->cpu.all.cycle = subcount(cur->cpu.all.cycle, pre->cpu.all.cycle);
+
 	for (i=0; i < dev->cpu.nrcpu; i++)
 	{
 		count_t 	ticks;
@@ -570,6 +628,11 @@ deviatsyst(struct sstat *cur, struct sstat *pre, struct sstat *dev, double inter
 		dev->cpu.cpu[i].guest = subcount(cur->cpu.cpu[i].guest,
 					         pre->cpu.cpu[i].guest);
 
+		dev->cpu.cpu[i].instr = subcount(cur->cpu.cpu[i].instr,
+						 pre->cpu.cpu[i].instr);
+		dev->cpu.cpu[i].cycle = subcount(cur->cpu.cpu[i].cycle,
+						 pre->cpu.cpu[i].cycle);
+
 		ticks 		      = cur->cpu.cpu[i].freqcnt.ticks;
 
 		dev->cpu.cpu[i].freqcnt.maxfreq = 
@@ -618,6 +681,22 @@ deviatsyst(struct sstat *cur, struct sstat *pre, struct sstat *dev, double inter
 	dev->mem.allocstall	= subcount(cur->mem.allocstall,
 				                         pre->mem.allocstall);
 
+	dev->psi          	= cur->psi;
+
+	if (cur->psi.present)
+	{
+		dev->psi.cpusome.total 	= cur->psi.cpusome.total -
+					  pre->psi.cpusome.total;
+		dev->psi.memsome.total 	= cur->psi.memsome.total -
+					  pre->psi.memsome.total;
+		dev->psi.memfull.total 	= cur->psi.memfull.total -
+					  pre->psi.memfull.total;
+		dev->psi.iosome.total 	= cur->psi.iosome.total -
+					  pre->psi.iosome.total;
+		dev->psi.iofull.total 	= cur->psi.iofull.total -
+					  pre->psi.iofull.total;
+	}
+
 	/*
 	** structures with network-related counters are considered
 	** as tables of frequency-counters that have to be subtracted;
@@ -1141,6 +1220,7 @@ deviatsyst(struct sstat *cur, struct sstat *pre, struct sstat *dev, double inter
                                    pre->nfs.nfsmounts.nfsmnt[j].pagesmwrite);
 	}
 
+	dev->nfs.nfsmounts.nfsmnt[i].mountdev[0] = '\0';
 	dev->nfs.nfsmounts.nrmounts = cur->nfs.nfsmounts.nrmounts;
 
 	/*
@@ -1148,7 +1228,7 @@ deviatsyst(struct sstat *cur, struct sstat *pre, struct sstat *dev, double inter
 	*/
 	if (cur->cfs.nrcontainer != dev->cfs.nrcontainer)
 	{
-		size = (cur->cfs.nrcontainer + 1) * sizeof(struct percontainer);
+		size = cur->cfs.nrcontainer * sizeof(struct percontainer);
 		dev->cfs.cont = (struct percontainer *)realloc(dev->cfs.cont, size);
 		ptrverify(dev->cfs.cont, "deviatsyst cont [%ld]\n", (long)size);
 	}
@@ -1200,6 +1280,85 @@ deviatsyst(struct sstat *cur, struct sstat *pre, struct sstat *dev, double inter
 
 	dev->cfs.nrcontainer = cur->cfs.nrcontainer;
 
+ 	/*
+	** application-specific counters
+	** calculate deviations for GPUs
+	*/
+	if (cur->gpu.nrgpus != dev->gpu.nrgpus)
+	{
+		size = cur->gpu.nrgpus * sizeof(struct pergpu);
+		dev->gpu.gpu = (struct pergpu *)realloc(dev->gpu.gpu, size);
+		ptrverify(dev->gpu.gpu, "deviatsyst gpu [%ld]\n", (long)size);
+	}
+	for (i=0; i < cur->gpu.nrgpus; i++)
+	{
+	    dev->gpu.gpu[i].gpunr      = i;
+
+	    strcpy(dev->gpu.gpu[i].type,  cur->gpu.gpu[i].type);
+	    strcpy(dev->gpu.gpu[i].busid, cur->gpu.gpu[i].busid);
+
+	    dev->gpu.gpu[i].taskstats  = cur->gpu.gpu[i].taskstats;
+	    dev->gpu.gpu[i].nrprocs    = cur->gpu.gpu[i].nrprocs;
+
+	    dev->gpu.gpu[i].gpupercnow = cur->gpu.gpu[i].gpupercnow;
+	    dev->gpu.gpu[i].mempercnow = cur->gpu.gpu[i].mempercnow;
+	    dev->gpu.gpu[i].memtotnow  = cur->gpu.gpu[i].memtotnow;
+	    dev->gpu.gpu[i].memusenow  = cur->gpu.gpu[i].memusenow;
+
+	    dev->gpu.gpu[i].samples    = subcount(cur->gpu.gpu[i].samples,
+	                                          pre->gpu.gpu[i].samples);
+
+	    if (cur->gpu.gpu[i].gpuperccum >= 0)
+	     dev->gpu.gpu[i].gpuperccum = subcount(cur->gpu.gpu[i].gpuperccum,
+	                                           pre->gpu.gpu[i].gpuperccum);
+	    else
+	     dev->gpu.gpu[i].gpuperccum = -1;
+
+	    if (cur->gpu.gpu[i].memusecum >= 0)
+	     dev->gpu.gpu[i].memusecum  = subcount(cur->gpu.gpu[i].memusecum,
+	                                           pre->gpu.gpu[i].memusecum);
+	    else
+	     dev->gpu.gpu[i].memusecum  = -1;
+
+	    if (cur->gpu.gpu[i].memperccum >= 0)
+	     dev->gpu.gpu[i].memperccum = subcount(cur->gpu.gpu[i].memperccum,
+	                                           pre->gpu.gpu[i].memperccum);
+	    else
+	     dev->gpu.gpu[i].memperccum = -1;
+	}
+
+	dev->gpu.nrgpus = cur->gpu.nrgpus;
+
+	/*
+	** calculate deviations for InfiniBand
+	*/
+	if (cur->ifb.nrports != dev->ifb.nrports)
+	{
+		size = (cur->ifb.nrports + 1) * sizeof(struct perifb);
+		dev->ifb.ifb = (struct perifb *)realloc(dev->ifb.ifb, size);
+		ptrverify(dev->ifb.ifb, "deviatsyst ifb [%ld]\n", (long)size);
+	}
+	for (i=0; i < cur->ifb.nrports; i++)
+	{
+		strcpy(dev->ifb.ifb[i].ibname, cur->ifb.ifb[i].ibname);
+
+		dev->ifb.ifb[i].portnr = cur->ifb.ifb[i].portnr;
+		dev->ifb.ifb[i].lanes  = cur->ifb.ifb[i].lanes;
+		dev->ifb.ifb[i].rate   = cur->ifb.ifb[i].rate;
+
+		dev->ifb.ifb[i].rcvb   = cur->ifb.ifb[i].rcvb -
+		                         pre->ifb.ifb[i].rcvb;
+		dev->ifb.ifb[i].sndb   = cur->ifb.ifb[i].sndb -
+		                         pre->ifb.ifb[i].sndb;
+		dev->ifb.ifb[i].rcvp   = cur->ifb.ifb[i].rcvp -
+		                         pre->ifb.ifb[i].rcvp;
+		dev->ifb.ifb[i].sndp   = cur->ifb.ifb[i].sndp -
+		                         pre->ifb.ifb[i].sndp;
+	}
+
+	dev->ifb.ifb[i].ibname[0] = '\0';
+	dev->ifb.nrports = cur->ifb.nrports;
+
 	/*
 	** application-specific counters
 	*/
@@ -1229,6 +1388,7 @@ totalsyst(char category, struct sstat *new, struct sstat *tot)
 {
 	register int	i;
 	count_t		*ctot, *cnew;
+	size_t		size;
 
 	switch (category)
 	{
@@ -1248,6 +1408,13 @@ totalsyst(char category, struct sstat *new, struct sstat *tot)
 		tot->cpu.all.steal += new->cpu.all.steal;
 		tot->cpu.all.guest += new->cpu.all.guest;
 
+		if (tot->cpu.nrcpu < new->cpu.nrcpu || !tot->cpu.cpu)
+		{
+		    size = new->cpu.nrcpu * sizeof(struct percpu);
+		    tot->cpu.cpu = realloc(tot->cpu.cpu, size);
+		    ptrverify(tot->cpu.cpu, "totalsyst cpus [%ld]", (long)size);
+		}
+
 		if (new->cpu.nrcpu == 1)
 		{
 			tot->cpu.cpu[0] = tot->cpu.all;
@@ -1392,6 +1559,13 @@ totalsyst(char category, struct sstat *new, struct sstat *tot)
 		tot->net.tcp.MaxConn      = new->net.tcp.MaxConn;
 		tot->net.tcp.CurrEstab    = new->net.tcp.CurrEstab;
 	
+		if (tot->intf.nrintf < new->intf.nrintf || !tot->intf.intf)
+		{
+		    size = (new->intf.nrintf + 1) * sizeof(struct perintf);
+		    tot->intf.intf = realloc(tot->intf.intf, size);
+		    ptrverify(tot->intf.intf, "totalsyst intfs [%ld]", (long)size);
+		}
+
 		for (i=0; new->intf.intf[i].name[0]; i++)
 		{
 			/*
@@ -1458,6 +1632,13 @@ totalsyst(char category, struct sstat *new, struct sstat *tot)
 		break;
 
 	   case 'd':	/* accumulate disk-related counters */
+		if (tot->dsk.ndsk < new->dsk.ndsk || !tot->dsk.dsk)
+		{
+		    size = (new->dsk.ndsk + 1) * sizeof(struct perdsk);
+		    tot->dsk.dsk = realloc(tot->dsk.dsk, size);
+		    ptrverify(tot->dsk.dsk, "totalsyst disks [%ld]", (long)size);
+		}
+
 		for (i=0; new->dsk.dsk[i].name[0]; i++)
 		{
 			strcpy(tot->dsk.dsk[i].name, new->dsk.dsk[i].name);
@@ -1473,6 +1654,13 @@ totalsyst(char category, struct sstat *new, struct sstat *tot)
 		tot->dsk.dsk[i].name[0] = '\0';
 		tot->dsk.ndsk = i;
 
+		if (tot->dsk.nlvm < new->dsk.nlvm || !tot->dsk.lvm)
+		{
+		    size = (new->dsk.nlvm + 1) * sizeof(struct perdsk);
+		    tot->dsk.lvm = realloc(tot->dsk.lvm, size);
+		    ptrverify(tot->dsk.lvm, "totalsyst LVs [%ld]", (long)size);
+		}
+
 		for (i=0; new->dsk.lvm[i].name[0]; i++)
 		{
 			strcpy(tot->dsk.lvm[i].name, new->dsk.lvm[i].name);
@@ -1488,6 +1676,13 @@ totalsyst(char category, struct sstat *new, struct sstat *tot)
 		tot->dsk.lvm[i].name[0] = '\0';
 		tot->dsk.nlvm = i;
 
+		if (tot->dsk.nmdd < new->dsk.nmdd || !tot->dsk.mdd)
+		{
+		    size = (new->dsk.nmdd + 1) * sizeof(struct perdsk);
+		    tot->dsk.mdd = realloc(tot->dsk.mdd, size);
+		    ptrverify(tot->dsk.mdd, "totalsyst MDs [%ld]", (long)size);
+		}
+
 		for (i=0; new->dsk.mdd[i].name[0]; i++)
 		{
 			strcpy(tot->dsk.mdd[i].name, new->dsk.mdd[i].name);
