@@ -18,6 +18,7 @@
 #include "atop.h"
 #include "photosyst.h"
 #include "systmetrics.h"
+#include "showgeneric.h"
 
 /*
 ** Allocate the fixes space for system statistics and the associated
@@ -274,7 +275,7 @@ photosyst(struct sstat *si)
 	}
 
 	if (fetch_metrics("system", SYST_NMETRICS, pmids, &result) < 0)
-		return 'r';
+		return MRESET;
 
 	onrcpu  = si->cpu.nrcpu;
 	onrintf = si->intf.nrintf;
@@ -296,6 +297,7 @@ photosyst(struct sstat *si)
 	si->cpu.csw = extract_count_t(result, descs, CPU_CSW);
 	si->cpu.devint = extract_count_t(result, descs, CPU_DEVINT);
 	si->cpu.nprocs = extract_count_t(result, descs, CPU_NPROCS);
+	si->cpu.btime  = extract_count_t(result, descs, CPU_BTIME);
 	si->cpu.all.utime = extract_count_t(result, descs, CPU_UTIME);
 	si->cpu.all.ntime = extract_count_t(result, descs, CPU_NTIME);
 	si->cpu.all.stime = extract_count_t(result, descs, CPU_STIME);
