@@ -1,6 +1,7 @@
 #!/usr/bin/env pmpython
 #
-# Copyright (C) 2015-2018 Marko Myllynen <myllynen@redhat.com>
+# Copyright (C) 2015-2019 Marko Myllynen <myllynen@redhat.com>
+# Copyright (C) 2014-2018 Red Hat.
 #
 # This program is free software; you can redistribute it and/or modify it
 # under the terms of the GNU General Public License as published by the
@@ -11,15 +12,9 @@
 # WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 # or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 # for more details.
-
-# [write_es] Copyright (C) 2014-2015 Red Hat, based on pcp2es by Frank Ch. Eigler
-
-# pylint: disable=superfluous-parens
-# pylint: disable=invalid-name, line-too-long, no-self-use
-# pylint: disable=too-many-boolean-expressions, too-many-statements
-# pylint: disable=too-many-instance-attributes, too-many-locals
-# pylint: disable=too-many-branches, too-many-nested-blocks
-# pylint: disable=broad-except
+#
+# pylint: disable=line-too-long, broad-except, bad-continuation
+#
 
 """ PCP to Elasticsearch Bridge """
 
@@ -30,12 +25,12 @@ import time
 import sys
 
 # Our imports
-from elasticsearch import Elasticsearch, ElasticsearchException
+import json
+import requests
 
 # PCP Python PMAPI
 from pcp import pmapi, pmconfig
-from cpmapi import PM_CONTEXT_ARCHIVE, PM_ERR_EOL, PM_IN_NULL, PM_DEBUG_APPL1
-from cpmapi import PM_TIME_MSEC
+from cpmapi import PM_CONTEXT_ARCHIVE, PM_IN_NULL, PM_DEBUG_APPL1, PM_TIME_MSEC
 
 if sys.version_info[0] >= 3:
     long = int # pylint: disable=redefined-builtin
@@ -45,8 +40,9 @@ DEFAULT_CONFIG = ["./pcp2elasticsearch.conf", "$HOME/.pcp2elasticsearch.conf", "
 
 # Defaults
 CONFVER = 1
-ES_INDEX = "pcp"
 ES_SERVER = "http://localhost:9200/"
+ES_INDEX = "pcp"
+ES_SEARCH_TYPE = "pcp-metric"
 
 class pcp2elasticsearch(object):
     """ PCP to Elasticsearch """
@@ -60,12 +56,13 @@ class pcp2elasticsearch(object):
         # Configuration directives
         self.keys = ('source', 'output', 'derived', 'header', 'globals',
                      'samples', 'interval', 'type', 'precision', 'daemonize',
-                     'es_server', 'es_index', 'es_hostid',
+                     'es_server', 'es_index', 'es_hostid', 'es_search_type',
                      'count_scale', 'space_scale', 'time_scale', 'version',
                      'count_scale_force', 'space_scale_force', 'time_scale_force',
                      'type_prefer', 'precision_force', 'limit_filter', 'limit_filter_force',
-                     'live_filter', 'rank', 'invert_filter', 'predicate',
-                     'speclocal', 'instances', 'ignore_incompat', 'omit_flat')
+                     'live_filter', 'rank', 'invert_filter', 'predicate', 'names_change',
+                     'speclocal', 'instances', 'ignore_incompat', 'ignore_unknown',
+                     'omit_flat')
 
         # The order of preference for options (as present):
         # 1 - command line options
@@ -86,6 +83,8 @@ class pcp2elasticsearch(object):
         self.type = 0
         self.type_prefer = self.type
         self.ignore_incompat = 0
+        self.ignore_unknown = 0
+        self.names_change = 0 # ignore
         self.instances = []
         self.live_filter = 0
         self.rank = 0
@@ -107,7 +106,9 @@ class pcp2elasticsearch(object):
 
         self.es_server = ES_SERVER
         self.es_index = ES_INDEX
+        self.es_search_type = ES_SEARCH_TYPE
         self.es_hostid = None
+        self.es_failed = False
 
         # Internal
         self.runtime = -1
@@ -132,7 +133,7 @@ class pcp2elasticsearch(object):
         opts = pmapi.pmOptions()
         opts.pmSetOptionCallback(self.option)
         opts.pmSetOverrideCallback(self.option_override)
-        opts.pmSetShortOptions("a:h:LK:c:Ce:D:V?HGA:S:T:O:s:t:rRIi:jJ:8:9:nN:vP:0:q:b:y:Q:B:Y:g:x:X:")
+        opts.pmSetShortOptions("a:h:LK:c:Ce:D:V?HGA:S:T:O:s:t:rRIi:jJ:4:58:9:nN:vP:0:q:b:y:Q:B:Y:g:x:X:p:")
         opts.pmSetShortUsage("[option...] metricspec [...]")
 
         opts.pmSetLongOptionHeader("General options")
@@ -145,7 +146,7 @@ class pcp2elasticsearch(object):
         opts.pmSetLongOption("config", 1, "c", "FILE", "config file path")
         opts.pmSetLongOption("check", 0, "C", "", "check config and metrics and exit")
         opts.pmSetLongOption("derived", 1, "e", "FILE|DFNT", "derived metrics definitions")
-        self.daemonize = opts.pmSetLongOption("daemonize", 0, "", "", "daemonize on startup") # > 1
+        opts.pmSetLongOption("daemonize", 0, "", "", "daemonize on startup")
         opts.pmSetLongOptionDebug()        # -D/--debug
         opts.pmSetLongOptionVersion()      # -V/--version
         opts.pmSetLongOptionHelp()         # -?/--help
@@ -162,6 +163,8 @@ class pcp2elasticsearch(object):
         opts.pmSetLongOption("raw", 0, "r", "", "output raw counter values (no rate conversion)")
         opts.pmSetLongOption("raw-prefer", 0, "R", "", "prefer output raw counter values (no rate conversion)")
         opts.pmSetLongOption("ignore-incompat", 0, "I", "", "ignore incompatible instances (default: abort)")
+        opts.pmSetLongOption("ignore-unknown", 0, "5", "", "ignore unknown metrics (default: abort)")
+        opts.pmSetLongOption("names-change", 1, "4", "ACTION", "update/ignore/abort on PMNS change (default: ignore)")
         opts.pmSetLongOption("instances", 1, "i", "STR", "instances to report (default: all current)")
         opts.pmSetLongOption("live-filter", 0, "j", "", "perform instance live filtering")
         opts.pmSetLongOption("rank", 1, "J", "COUNT", "limit results to COUNT highest/lowest valued instances")
@@ -179,9 +182,10 @@ class pcp2elasticsearch(object):
         opts.pmSetLongOption("time-scale", 1, "y", "SCALE", "default time unit")
         opts.pmSetLongOption("time-scale-force", 1, "Y", "SCALE", "forced time unit")
 
-        opts.pmSetLongOption("es-host", 1, "g", "SERVER", "elasticsearch server (default: " + ES_SERVER + ")")
-        opts.pmSetLongOption("es-index", 1, "x", "INDEX", "elasticsearch index for metric names (default: " + ES_INDEX + ")")
-        opts.pmSetLongOption("es-hostid", 1, "X", "HOSTID", "elasticsearch host-id for measurements")
+        opts.pmSetLongOption("es-host", 1, "g", "SERVER", "Elasticsearch server (default: " + ES_SERVER + ")")
+        opts.pmSetLongOption("es-index", 1, "x", "INDEX", "Elasticsearch index for metric names (default: " + ES_INDEX + ")")
+        opts.pmSetLongOption("es-hostid", 1, "X", "HOSTID", "Elasticsearch host-id for measurements")
+        opts.pmSetLongOption("es-search-type", 1, "p", "TYPE", "Elasticsearch search type for measurements")
 
         return opts
 
@@ -191,12 +195,11 @@ class pcp2elasticsearch(object):
             return 1
         return 0
 
-    def option(self, opt, optarg, index):
-        """ Perform setup for an individual command line option """
-        if index == self.daemonize and opt == '':
+    def option(self, opt, optarg, _index):
+        """ Perform setup for individual command line option """
+        if opt == 'daemonize':
             self.daemonize = 1
-            return
-        if opt == 'K':
+        elif opt == 'K':
             if not self.speclocal or not self.speclocal.startswith(";"):
                 self.speclocal = ";" + optarg
             else:
@@ -220,6 +223,18 @@ class pcp2elasticsearch(object):
             self.type_prefer = 1
         elif opt == 'I':
             self.ignore_incompat = 1
+        elif opt == '5':
+            self.ignore_unknown = 1
+        elif opt == '4':
+            if optarg == 'ignore':
+                self.names_change = 0
+            elif optarg == 'abort':
+                self.names_change = 1
+            elif optarg == 'update':
+                self.names_change = 2
+            else:
+                sys.stderr.write("Unknown names-change action '%s' specified.\n" % optarg)
+                sys.exit(1)
         elif opt == 'i':
             self.instances = self.instances + self.pmconfig.parse_instances(optarg)
         elif opt == 'j':
@@ -258,11 +273,13 @@ class pcp2elasticsearch(object):
             self.es_index = optarg
         elif opt == 'X':
             self.es_hostid = optarg
+        elif opt == 'p':
+            self.es_search_type = optarg
         else:
             raise pmapi.pmUsageErr()
 
     def connect(self):
-        """ Establish a PMAPI context """
+        """ Establish PMAPI context """
         context, self.source = pmapi.pmContext.set_connect_options(self.opts, self.source, self.speclocal)
 
         self.pmfg = pmapi.fetchgroup(context, self.source)
@@ -320,19 +337,17 @@ class pcp2elasticsearch(object):
             time.sleep(align)
 
         # Main loop
+        refresh_metrics = 0
         while self.samples != 0:
-            # Fetch values
-            try:
-                self.pmfg.fetch()
-            except pmapi.pmErr as error:
-                if error.args[0] == PM_ERR_EOL:
-                    break
-                raise error
+            # Refresh metrics as needed
+            if refresh_metrics:
+                refresh_metrics = 0
+                self.pmconfig.update_metrics(curr_insts=not self.live_filter)
 
-            # Watch for endtime in uninterpolated mode
-            if not self.interpol:
-                if float(self.pmfg_ts().strftime('%s')) > float(self.opts.pmGetOptionFinish()):
-                    break
+            # Fetch values
+            refresh_metrics = self.pmconfig.fetch()
+            if refresh_metrics < 0:
+                break
 
             # Report and prepare for the next round
             self.report(self.pmfg_ts())
@@ -345,8 +360,8 @@ class pcp2elasticsearch(object):
         self.report(None)
 
     def report(self, tstamp):
-        """ Report the metric values """
-        if tstamp != None:
+        """ Report metric values """
+        if tstamp is not None:
             tstamp = tstamp.strftime(self.timefmt)
 
         self.write_es(tstamp)
@@ -357,7 +372,7 @@ class pcp2elasticsearch(object):
             sys.stdout.write("Sending %d archived metrics to Elasticsearch at %s...\n(Ctrl-C to stop)\n" % (len(self.metrics), self.es_server))
             return
 
-        sys.stdout.write("Sending %d metrics to Elasticsearch at %s every %d sec" % (len(self.metrics), self.es_server, self.interval))
+        sys.stdout.write("Sending %d metrics to Elasticsearch at %s every %.1f sec" % (len(self.metrics), self.es_server, float(self.interval)))
         if self.runtime != -1:
             sys.stdout.write(":\n%s samples(s) with %.1f sec interval ~ %d sec runtime.\n" % (self.samples, float(self.interval), self.runtime))
         elif self.samples:
@@ -367,7 +382,7 @@ class pcp2elasticsearch(object):
             sys.stdout.write("...\n(Ctrl-C to stop)\n")
 
     def write_es(self, timestamp):
-        """ Write (send) metrics to an Elasticsearch host """
+        """ Write (send) metrics to Elasticsearch host """
         if timestamp is None:
             # Silent goodbye
             return
@@ -375,15 +390,20 @@ class pcp2elasticsearch(object):
         ts = self.context.datetime_to_secs(self.pmfg_ts(), PM_TIME_MSEC)
 
         try:
-            es = Elasticsearch(hosts=[self.es_server])
-            # pylint: disable=unexpected-keyword-arg
-            es.indices.create(index=self.es_index,
-                              ignore=[400],
-                              body={'mappings':{'pcp-metric':
-                                                {'properties':{'@timestamp':{'type':'date'},
-                                                               'host-id':{'type':'string'}}}}})
-        except ElasticsearchException as error:
-            sys.stderr.write("Can't connect to Elasticsearch server %s: %s, continuing.\n" % (self.es_server, str(error)))
+            body = {'ignore': 400,
+                    'mappings': {'pcp-metric':
+                                {'properties':{'@timestamp':{'type':'epoch_milli'},
+                                               '@host-id':{'type':'string'}}}}}
+            headers = {'content-type': 'application/json'} # Do we need this?
+            url = self.es_server+'/'+self.es_index
+            requests.put(url, data=json.dumps(body), headers=headers)
+            if self.es_failed:
+                sys.stderr.write("Reconnected to Elasticsearch server %s.\n" % (self.es_server))
+            self.es_failed = False
+        except Exception as error:
+            if not self.es_failed:
+                sys.stderr.write("Can't connect to Elasticsearch server %s: %s, continuing.\n" % (self.es_server, str(error)))
+            self.es_failed = True
             return
 
         # Assemble all metrics into a single document
@@ -393,7 +413,7 @@ class pcp2elasticsearch(object):
         insts_key = "@instances"
         inst_key = "@id"
 
-        results = self.pmconfig.get_sorted_results()
+        results = self.pmconfig.get_ranked_results(valid_only=True)
 
         for metric in results:
             # Install value into outgoing json/dict in key1{key2{key3=value}} style:
@@ -429,13 +449,10 @@ class pcp2elasticsearch(object):
                         insts.append({inst_key: name, last_part: value})
 
         try:
-            # pylint: disable=unexpected-keyword-arg
-            es.index(index=self.es_index,
-                     doc_type='pcp-metric',
-                     timestamp=long(ts),
-                     body=es_doc)
-        except ElasticsearchException as error:
-            sys.stderr.write("Can't send to Elasticsearch server %s: %s, continuing.\n" % (self.es_server, str(error)))
+            url = self.es_server + '/' + self.es_index + '/' + self.es_search_type
+            requests.post(url, data=json.dumps(es_doc), headers=headers)
+        except Exception as error:
+            sys.stderr.write("Cannot send to Elasticsearch server %s: %s, continuing.\n" % (self.es_server, str(error)))
             return
 
     def finalize(self):
@@ -449,9 +466,11 @@ if __name__ == '__main__':
         P.validate_config()
         P.execute()
         P.finalize()
-
     except pmapi.pmErr as error:
-        sys.stderr.write('%s: %s\n' % (error.progname(), error.message()))
+        sys.stderr.write("%s: %s" % (error.progname(), error.message()))
+        if error.message() == "Connection refused":
+            sys.stderr.write("; is pmcd running?")
+        sys.stderr.write("\n")
         sys.exit(1)
     except pmapi.pmUsageErr as usage:
         usage.message()
