@@ -1,6 +1,6 @@
 #!/usr/bin/env pmpython
 #
-# Copyright (C) 2015-2018 Marko Myllynen <myllynen@redhat.com>
+# Copyright (C) 2015-2019 Marko Myllynen <myllynen@redhat.com>
 # Copyright (C) 2018 Red Hat.
 #
 # This program is free software; you can redistribute it and/or modify it
@@ -34,7 +34,7 @@ import json
 
 # PCP Python PMAPI
 from pcp import pmapi, pmconfig
-from cpmapi import PM_CONTEXT_ARCHIVE, PM_ERR_EOL, PM_IN_NULL, PM_DEBUG_APPL1
+from cpmapi import PM_CONTEXT_ARCHIVE, PM_IN_NULL, PM_DEBUG_APPL0, PM_DEBUG_APPL1
 from cpmapi import PM_TIME_MSEC
 
 if sys.version_info[0] >= 3:
@@ -45,13 +45,10 @@ DEFAULT_CONFIG = ["./pcp2spark.conf", "$HOME/.pcp2spark.conf", "$HOME/.pcp/pcp2s
 
 # Defaults
 CONFVER = 1
-INDENT = 2
-
-# Spark server
 SPARK_SERVER = "127.0.0.1"
-SPARK_PORT = "44325"
+SPARK_PORT = 44325
 
-class PCP2SPARK(object):
+class PCP2Spark(object):
     """ PCP to Spark """
     def __init__(self):
         """ Construct object, prepare for command line handling """
@@ -67,8 +64,9 @@ class PCP2SPARK(object):
                      'count_scale', 'space_scale', 'time_scale', 'version',
                      'count_scale_force', 'space_scale_force', 'time_scale_force',
                      'type_prefer', 'precision_force', 'limit_filter', 'limit_filter_force',
-                     'live_filter', 'rank', 'invert_filter', 'predicate',
-                     'speclocal', 'instances', 'ignore_incompat', 'omit_flat')
+                     'live_filter', 'rank', 'invert_filter', 'predicate', 'names_change',
+                     'speclocal', 'instances', 'ignore_incompat', 'ignore_unknown',
+                     'omit_flat')
 
         # The order of preference for options (as present):
         # 1 - command line options
@@ -89,6 +87,8 @@ class PCP2SPARK(object):
         self.type = 0
         self.type_prefer = self.type
         self.ignore_incompat = 0
+        self.ignore_unknown = 0
+        self.names_change = 0 # ignore
         self.instances = []
         self.live_filter = 0
         self.rank = 0
@@ -113,9 +113,8 @@ class PCP2SPARK(object):
 
         # Internal
         self.runtime = -1
-        self.spark_socket = None
+
         self.client_conn = None
-        self.address = None
 
         # Performance metrics store
         # key - metric name
@@ -137,7 +136,7 @@ class PCP2SPARK(object):
         opts = pmapi.pmOptions()
         opts.pmSetOptionCallback(self.option)
         opts.pmSetOverrideCallback(self.option_override)
-        opts.pmSetShortOptions("a:h:LK:c:Ce:D:V?HGA:S:T:O:s:t:rRIi:jJ:8:9:nN:vP:0:q:b:y:Q:B:Y:g:p:")
+        opts.pmSetShortOptions("a:h:LK:c:Ce:D:V?HGA:S:T:O:s:t:rRIi:jJ:4:58:9:nN:vP:0:q:b:y:Q:B:Y:g:p:")
         opts.pmSetShortUsage("[option...] metricspec [...]")
 
         opts.pmSetLongOptionHeader("General options")
@@ -167,6 +166,8 @@ class PCP2SPARK(object):
         opts.pmSetLongOption("raw", 0, "r", "", "output raw counter values (no rate conversion)")
         opts.pmSetLongOption("raw-prefer", 0, "R", "", "prefer output raw counter values (no rate conversion)")
         opts.pmSetLongOption("ignore-incompat", 0, "I", "", "ignore incompatible instances (default: abort)")
+        opts.pmSetLongOption("ignore-unknown", 0, "5", "", "ignore unknown metrics (default: abort)")
+        opts.pmSetLongOption("names-change", 1, "4", "ACTION", "update/ignore/abort on PMNS change (default: ignore)")
         opts.pmSetLongOption("instances", 1, "i", "STR", "instances to report (default: all current)")
         opts.pmSetLongOption("live-filter", 0, "j", "", "perform instance live filtering")
         opts.pmSetLongOption("rank", 1, "J", "COUNT", "limit results to COUNT highest/lowest valued instances")
@@ -184,8 +185,8 @@ class PCP2SPARK(object):
         opts.pmSetLongOption("time-scale", 1, "y", "SCALE", "default time unit")
         opts.pmSetLongOption("time-scale-force", 1, "Y", "SCALE", "forced time unit")
 
-        opts.pmSetLongOption("spark_server", 1, "g", "SERVER", "IP address for local server for metrics (default: " + str(SPARK_SERVER) + ")")
-        opts.pmSetLongOption("spark_port", 1, "p", "PORT", "Port for local server for metrics (default: " + str(SPARK_PORT) + ")")
+        opts.pmSetLongOption("spark-server", 1, "g", "SERVER", "Spark server (default: " + SPARK_SERVER + ")")
+        opts.pmSetLongOption("spark-port", 1, "p", "PORT", "Spark port (default: " + str(SPARK_PORT) + ")")
 
         return opts
 
@@ -195,8 +196,8 @@ class PCP2SPARK(object):
             return 1
         return 0
 
-    def option(self, opt, optarg, index):
-        """ Perform setup for an individual command line option """
+    def option(self, opt, optarg, _index):
+        """ Perform setup for individual command line option """
         if opt == 'daemonize':
             self.daemonize = 1
         elif opt == 'K':
@@ -223,6 +224,18 @@ class PCP2SPARK(object):
             self.type_prefer = 1
         elif opt == 'I':
             self.ignore_incompat = 1
+        elif opt == '5':
+            self.ignore_unknown = 1
+        elif opt == '4':
+            if optarg == 'ignore':
+                self.names_change = 0
+            elif optarg == 'abort':
+                self.names_change = 1
+            elif optarg == 'update':
+                self.names_change = 2
+            else:
+                sys.stderr.write("Unknown names-change action '%s' specified.\n" % optarg)
+                sys.exit(1)
         elif opt == 'i':
             self.instances = self.instances + self.pmconfig.parse_instances(optarg)
         elif opt == 'j':
@@ -258,12 +271,12 @@ class PCP2SPARK(object):
         elif opt == 'g':
             self.spark_server = optarg
         elif opt == 'p':
-            self.spark_port = optarg
+            self.spark_port = int(optarg)
         else:
             raise pmapi.pmUsageErr()
 
     def connect(self):
-        """ Establish a PMAPI context """
+        """ Establish PMAPI context """
         context, self.source = pmapi.pmContext.set_connect_options(self.opts, self.source, self.speclocal)
 
         self.pmfg = pmapi.fetchgroup(context, self.source)
@@ -299,14 +312,14 @@ class PCP2SPARK(object):
         self.context.prepare_execute(self.opts, False, self.interpol, self.interval)
 
         # Setup a socket to send data to Spark
-        self.spark_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-        self.spark_socket.bind((self.spark_server, int(self.spark_port)))
-        self.spark_socket.listen(0) # Don't handle listen backlog, only allow single connection
+        spark_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+        spark_socket.bind((self.spark_server, self.spark_port))
+        spark_socket.listen(0) # Don't handle listen backlog, only allow single connection
 
-        # At this point we wait for a connection from the spark worker job
-        # and when we see it accept the connection.
+        # Wait for a connection from the Spark worker
+        # job and when we see it accept the connection
         sys.stdout.write("Awaiting connection...\n")
-        self.client_conn, self.address = self.spark_socket.accept()
+        self.client_conn, _ = spark_socket.accept()
 
         # Headers
         if self.header == 1:
@@ -327,20 +340,17 @@ class PCP2SPARK(object):
             time.sleep(align)
 
         # Main loop
+        refresh_metrics = 0
         while self.samples != 0:
+            # Refresh metrics as needed
+            if refresh_metrics:
+                refresh_metrics = 0
+                self.pmconfig.update_metrics(curr_insts=not self.live_filter)
 
             # Fetch values
-            try:
-                self.pmfg.fetch()
-            except pmapi.pmErr as error:
-                if error.args[0] == PM_ERR_EOL:
-                    break
-                raise error
-
-            # Watch for endtime in uninterpolated mode
-            if not self.interpol:
-                if float(self.pmfg_ts().strftime('%s')) > float(self.opts.pmGetOptionFinish()):
-                    break
+            refresh_metrics = self.pmconfig.fetch()
+            if refresh_metrics < 0:
+                break
 
             # Report and prepare for the next round
             self.report(self.pmfg_ts())
@@ -353,8 +363,8 @@ class PCP2SPARK(object):
         self.report(None)
 
     def report(self, tstamp):
-        """ Report the metric values """
-        if tstamp != None:
+        """ Report metric values """
+        if tstamp is not None:
             tstamp = tstamp.strftime(self.timefmt)
 
         self.write_spark(tstamp)
@@ -362,21 +372,21 @@ class PCP2SPARK(object):
     def write_header(self):
         """ Write info header """
         if self.context.type == PM_CONTEXT_ARCHIVE:
-            sys.stdout.write("Broadcasting %d archived metrics to Spark from %s:%s...\n(Ctrl-C to stop)\n" % (len(self.metrics), self.spark_server, self.spark_port))
+            sys.stdout.write("Sending %d archived metrics to Spark %s:%d...\n(Ctrl-C to stop)\n" % (len(self.metrics), self.spark_server, self.spark_port))
             return
 
-        sys.stdout.write("Broadcasting %d metrics to Spark from %s:%s every %d sec" % (len(self.metrics), self.spark_server, self.spark_port, self.interval))
+        sys.stdout.write("Sending %d metrics to Spark %s:%d every %.1f sec" % (len(self.metrics), self.spark_server, self.spark_port, float(self.interval)))
         if self.runtime != -1:
             sys.stdout.write(":\n%s samples(s) with %.1f sec interval ~ %d sec runtime.\n" % (self.samples, float(self.interval), self.runtime))
         elif self.samples:
             duration = (self.samples - 1) * float(self.interval)
             sys.stdout.write(":\n%s samples(s) with %.1f sec interval ~ %d sec runtime.\n" % (self.samples, float(self.interval), duration))
         else:
-            sys.stdout.write("...\n\n(Ctrl-C to stop)\n")
+            sys.stdout.write("...\n(Ctrl-C to stop)\n")
 
-    def write_spark(self, tstamp):
-        """ Write (send) metrics to socket for Spark sever to pickup """
-        if tstamp is None:
+    def write_spark(self, timestamp):
+        """ Write (send) metrics to socket for Spark server to pick up """
+        if timestamp is None:
             # Silent goodbye, close in finalize()
             return
 
@@ -390,7 +400,7 @@ class PCP2SPARK(object):
         insts_key = "@instances"
         inst_key = "@id"
 
-        results = self.pmconfig.get_sorted_results()
+        results = self.pmconfig.get_ranked_results(valid_only=True)
 
         for metric in results:
             # Install value into outgoing json/dict in key1{key2{key3=value}} style:
@@ -432,42 +442,40 @@ class PCP2SPARK(object):
                              ensure_ascii=False,
                              separators=(',', ': '))
 
+        if self.context.pmDebug(PM_DEBUG_APPL0):
+            print(message)
+
         try:
             self.client_conn.sendall(message.encode('utf-8'))
             self.client_conn.sendall("\n".encode('utf-8'))
-
         except socket.timeout as err:
-            sys.stderr.write("Connection to Spark timed out: " + str(err) + "\n")
+            sys.stderr.write("Spark connection timed out: %s\n" % str(err))
             sys.exit(1)
         except socket.error as err:
-            sys.stderr.write("Socket error: " + str(err) + "\n")
+            sys.stderr.write("Socket error: %s\n" % str(err))
             sys.exit(1)
 
-        # Test write output to console in debug
-        if self.context.pmDebug(PM_DEBUG_APPL1):
-            sys.stdout.write(message + "\n")
-
     def finalize(self):
         """ Finalize and clean up """
-        # Try and cleanly close socket
         if self.client_conn:
             try:
                 self.client_conn.close()
-            except: # pylint: disable=bare-except
+            except Exception:
                 pass
             self.client_conn = None
-        return
 
 if __name__ == '__main__':
     try:
-        P = PCP2SPARK()
+        P = PCP2Spark()
         P.connect()
         P.validate_config()
         P.execute()
         P.finalize()
-
     except pmapi.pmErr as error:
-        sys.stderr.write('%s: %s\n' % (error.progname(), error.message()))
+        sys.stderr.write("%s: %s" % (error.progname(), error.message()))
+        if error.message() == "Connection refused":
+            sys.stderr.write("; is pmcd running?")
+        sys.stderr.write("\n")
         sys.exit(1)
     except pmapi.pmUsageErr as usage:
         usage.message()
