@@ -1,7 +1,7 @@
 /*
  * Linux sysfs_kernel cluster
  *
- * Copyright (c) 2009, Red Hat, Inc.  All Rights Reserved.
+ * Copyright (c) 2009,2014,2016 Red Hat.
  * 
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -12,32 +12,30 @@
  * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  * for more details.
- * 
- * You should have received a copy of the GNU General Public License along
- * with this program; if not, write to the Free Software Foundation, Inc.,
- * 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
  */
-
+#include <ctype.h>
+#include "linux.h"
 #include "sysfs_kernel.h"
 
 int
 refresh_sysfs_kernel(sysfs_kernel_t *sk)
 {
-    char buf[64];
-    int fd;
+    char buf[MAXPATHLEN];
+    int fd, n;
 
-    if ((fd = open("/sys/kernel/uevent_seqnum", O_RDONLY)) < 0) {
+    pmsprintf(buf, sizeof(buf), "%s/sys/kernel/uevent_seqnum", linux_statspath);
+    if ((fd = open(buf, O_RDONLY)) < 0) {
     	sk->valid_uevent_seqnum = 0;
-	return -errno;
+	return -oserror();
     }
 
-    if (read(fd, buf, sizeof(buf)) <= 0)
+    if ((n = read(fd, buf, sizeof(buf))) <= 0)
     	sk->valid_uevent_seqnum = 0;
     else {
+	buf[n-1] = '\0';
     	sscanf(buf, "%llu", (long long unsigned int *)&sk->uevent_seqnum);
 	sk->valid_uevent_seqnum = 1;
     }
     close(fd);
-
     return 0;
 }
