@@ -1,6 +1,6 @@
 #! /bin/sh
 #
-# Copyright (c) 2013-2016,2018 Red Hat.
+# Copyright (c) 2013-2016,2018,2020 Red Hat.
 # Copyright (c) 1998-2000,2003 Silicon Graphics, Inc.  All Rights Reserved.
 # 
 # This program is free software; you can redistribute it and/or modify it
@@ -47,21 +47,29 @@ _unsymlink_path()
 
 # constant setup
 #
-tmp=`mktemp -d /tmp/pcp.XXXXXXXXX` || exit 1
+tmp=`mktemp -d "$PCP_TMPFILE_DIR/pmie_check.XXXXXXXXX"` || exit 1
 status=0
 echo >$tmp/lock
 prog=`basename $0`
 PROGLOG=$PCP_LOG_DIR/pmie/$prog.log
+MYPROGLOG=$PROGLOG.$$
 USE_SYSLOG=true
 
 _cleanup()
 {
+    if [ -s "$MYPROGLOG" ]
+    then
+	rm -f "$PROGLOG"
+	mv "$MYPROGLOG" "$PROGLOG"
+    else
+	rm -f "$MYPROGLOG"
+    fi
     $USE_SYSLOG && [ $status -ne 0 ] && \
     $PCP_SYSLOG_PROG -p daemon.error "$prog failed - see $PROGLOG"
-    [ -s "$PROGLOG" ] || rm -f "$PROGLOG"
     lockfile=`cat $tmp/lock 2>/dev/null`
     rm -f "$lockfile"
     rm -rf $tmp
+    $VERY_VERBOSE && echo "End: `date '+%F %T.%N'`"
 }
 trap "_cleanup; exit \$status" 0 1 2 3 15
 
@@ -126,6 +134,7 @@ do
 	-C)	CHECK_RUNLEVEL=true
 		;;
 	-l)	PROGLOG="$2"
+		MYPROGLOG="$PROGLOG".$$
 		USE_SYSLOG=false
 		shift
 		;;
@@ -187,11 +196,21 @@ else
     #
     # Exception ($SHOWME, above) is for -N where we want to see the output.
     #
-    touch "$PROGLOG"
-    chown $PCP_USER:$PCP_GROUP "$PROGLOG" >/dev/null 2>&1
-    exec 1>"$PROGLOG" 2>&1
+    touch "$MYPROGLOG"
+    chown $PCP_USER:$PCP_GROUP "$MYPROGLOG" >/dev/null 2>&1
+    exec 1>"$MYPROGLOG" 2>&1
 fi
 
+if $VERY_VERBOSE
+then
+    echo "Start: `date '+%F %T.%N'`"
+    if which pstree >/dev/null 2>&1
+    then
+	echo "Called from:"
+	pstree -spa $$
+	echo "--- end of pstree output ---"
+    fi
+fi
 _error()
 {
     echo "$prog: [$controlfile:$line]"
@@ -726,9 +745,9 @@ NR == 3	{ printf "p_pmcd_host=\"%s\"\n", $0; next }
 		_configure_pmie "$configfile" "$host" "$primary"
 	    fi
 
-	    if [ "X$primary" = Xy ]
+	    if [ "$primary" = y ]
 	    then
-		args="-P -l $logfile $args"
+		args="-F -P -l $logfile $args"
 	    else
 		args="-h $host -l $logfile $args"
 	    fi
@@ -767,13 +786,12 @@ NR == 3	{ printf "p_pmcd_host=\"%s\"\n", $0; next }
 		_unlock
 		continue
 	    else
-		# since this is launched as a sort of daemon, any output should
+		# since this is launched as a daemon, any output should
 		# go on pmie's stderr, i.e. $logfile ... use -b for this
 		#
 		$VERY_VERBOSE && ( echo; $PCP_ECHO_PROG $PCP_ECHO_N "+ ${sock_me}$PMIE -b $args""$PCP_ECHO_C"; echo "..." )
 		$PCP_BINADM_DIR/pmpost "start pmie from $prog for host $host"
-		${sock_me}$PMIE -b $args &
-		pid=$!
+		pid=`${sock_me} $PMIE -b $args >/dev/null 2>&1 & echo $!`
 	    fi
 
 	    # wait for pmie to get started, and check on its health
