@@ -43,7 +43,7 @@ unset PCP_STDERR
 
 # constant setup
 #
-tmp=`mktemp -d /tmp/pcp.XXXXXXXXX` || exit 1
+tmp=`mktemp -d "$PCP_TMPFILE_DIR/pmie_daily.XXXXXXXXX"` || exit 1
 status=0
 echo >$tmp/lock
 prog=`basename $0`
