@@ -31,21 +31,29 @@ unset PCP_STDERR
 
 # constant setup
 #
-tmp=`mktemp -d /tmp/pcp.XXXXXXXXX` || exit 1
+tmp=`mktemp -d "$PCP_TMPFILE_DIR/pmlogger_check.XXXXXXXXX"` || exit 1
 status=0
 echo >$tmp/lock
 prog=`basename $0`
 PROGLOG=$PCP_LOG_DIR/pmlogger/$prog.log
+MYPROGLOG=$PROGLOG.$$
 USE_SYSLOG=true
 
 _cleanup()
 {
+    if [ -s "$MYPROGLOG" ]
+    then
+	rm -f "$PROGLOG"
+	mv "$MYPROGLOG" "$PROGLOG"
+    else
+	rm -f "$MYPROGLOG"
+    fi
     $USE_SYSLOG && [ $status -ne 0 ] && \
     $PCP_SYSLOG_PROG -p daemon.error "$prog failed - see $PROGLOG"
-    [ -s "$PROGLOG" ] || rm -f "$PROGLOG"
     lockfile=`cat $tmp/lock 2>/dev/null`
     rm -f "$lockfile"
     rm -rf $tmp
+    $VERY_VERBOSE && echo "End: `date '+%F %T.%N'`"
 }
 trap "_cleanup; exit \$status" 0 1 2 3 15
 
@@ -86,6 +94,8 @@ VERY_VERBOSE=false
 CHECK_RUNLEVEL=false
 START_PMLOGGER=true
 STOP_PMLOGGER=false
+QUICKSTART=false
+SKIP_PRIMARY=false
 
 echo > $tmp/usage
 cat >> $tmp/usage << EOF
@@ -94,6 +104,8 @@ Options:
   -l=FILE,--logfile=FILE  send important diagnostic messages to FILE
   -C                      query system service runlevel information
   -N,--showme             perform a dry run, showing what would be done
+  -p,--skip-primary       do not start or stop the primary pmlogger instance
+  -q,--quick              quick start, no compression
   -s,--stop               stop pmlogger processes instead of starting them
   -T,--terse              produce a terser form of output
   -V,--verbose            increase diagnostic verbosity
@@ -117,6 +129,7 @@ do
 	-C)	CHECK_RUNLEVEL=true
 		;;
 	-l)	PROGLOG="$2"
+		MYPROGLOG="$PROGLOG".$$
 		USE_SYSLOG=false
 		daily_args="${daily_args} -l $2.from.check"
 		shift
@@ -129,6 +142,10 @@ do
 		KILL="echo + kill"
 		daily_args="${daily_args} -N"
 		;;
+	-p)	SKIP_PRIMARY=true
+		;;
+	-q)	QUICKSTART=true
+		;;
 	-s)	START_PMLOGGER=false
 		STOP_PMLOGGER=true
 		;;
@@ -162,9 +179,15 @@ fi
 
 _compress_now()
 {
-    # If $PCP_COMPRESSAFTER=0 in the control file(s), compress archives now.
-    # Invoked just before exit when this script has finished successfully.
-    $PCP_BINADM_DIR/pmlogger_daily -K $daily_args
+    if $QUICKSTART
+    then
+	$VERY_VERBOSE && echo "Skip compression, -q/--quick on command line"
+    else
+	# If $PCP_COMPRESSAFTER=0 in the control file(s), compress archives now.
+	# Invoked just before exit when this script has finished successfully.
+	$VERY_VERBOSE && echo "Doing compression ..."
+	$PCP_BINADM_DIR/pmlogger_daily -K $daily_args
+    fi
 }
 
 # after argument checking, everything must be logged to ensure no mail is
@@ -187,26 +210,37 @@ else
     #
     # Exception ($SHOWME, above) is for -N where we want to see the output.
     #
-    touch "$PROGLOG"
-    chown $PCP_USER:$PCP_GROUP "$PROGLOG" >/dev/null 2>&1
-    exec 1>"$PROGLOG" 2>&1
+    touch "$MYPROGLOG"
+    chown $PCP_USER:$PCP_GROUP "$MYPROGLOG" >/dev/null 2>&1
+    exec 1>"$MYPROGLOG" 2>&1
+fi
+
+if $VERY_VERBOSE
+then
+    echo "Start: `date '+%F %T.%N'`"
+    if which pstree >/dev/null 2>&1
+    then
+	echo "Called from:"
+	pstree -spa $$
+	echo "--- end of pstree output ---"
+    fi
 fi
 
 # if SaveLogs exists in the $PCP_LOG_DIR/pmlogger directory then save
-# $PROGLOG there as well with a unique name that contains the date and time
+# $MYPROGLOG there as well with a unique name that contains the date and time
 # when we're run
 #
 if [ -d $PCP_LOG_DIR/pmlogger/SaveLogs ]
 then
-    now="`date '+%Y%m%d.%H.%M'`"
-    link=`echo $PROGLOG | sed -e "s/$prog/SaveLogs\/$prog.$now/"`
+    now="`date '+%Y%m%d.%H.%M.%S'`"
+    link=`echo $MYPROGLOG | sed -e "s/$prog/SaveLogs\/$prog.$now/"`
     if [ ! -f "$link" ]
     then
 	if $SHOWME
 	then
-	    echo "+ ln $PROGLOG $link"
+	    echo "+ ln $MYPROGLOG $link"
 	else
-	    ln $PROGLOG $link
+	    ln $MYPROGLOG $link
 	fi
     fi
 fi
@@ -273,7 +307,7 @@ _restarting()
 
 _unlock()
 {
-    rm -f lock
+    rm -f "$1/lock"
     echo >$tmp/lock
 }
 
@@ -307,6 +341,8 @@ _configure_pmlogger()
 	grep '^# Auto-generated by pmlogconf' "$configfile" >/dev/null
 	owned=$?
 	skip=0
+	# assign PMLOGGER_CHECK_SKIP_LOGCONF (if present) from $envs
+	[ -n "$envs" -a -z "$PMLOGGER_CHECK_SKIP_LOGCONF" ] && eval $envs
 	if [ -n "$PMLOGGER_CHECK_SKIP_LOGCONF" ]
 	then
 	    # $PMLOGGER_CHECK_SKIP_LOGCONF is primarily for QA.
@@ -320,7 +356,7 @@ _configure_pmlogger()
 	then
 	    # pmlogconf file that we own, see if re-generation is needed
 	    cp "$configfile" "$tmpconfig"
-	    if $PMLOGCONF -r -c -q -h $hostname "$tmpconfig" >$tmp/diag 2>&1
+	    if $PMLOGCONF -r -c -q -h $hostname "$tmpconfig" </dev/null >$tmp/diag 2>&1
 	    then
 		if grep 'No changes' $tmp/diag >/dev/null 2>&1
 		then
@@ -328,6 +364,7 @@ _configure_pmlogger()
 		elif [ -w "$configfile" ]
 		then
 		    $VERBOSE && echo "Reconfigured: \"$configfile\" (pmlogconf)"
+		    chown $PCP_USER:$PCP_GROUP "$tmpconfig" >/dev/null 2>&1
 		    eval $MV "$tmpconfig" "$configfile"
 		else
 		    _warning "no write access to pmlogconf file \"$configfile\", skip reconfiguration"
@@ -340,6 +377,7 @@ _configure_pmlogger()
 		cat "$tmpconfig"
 		echo "=== end pmlogconf file ==="
 	    fi
+	    rm -f "$tmpconfig"
 	fi
     elif [ ! -e "$configfile" ]
     then
@@ -347,7 +385,7 @@ _configure_pmlogger()
 	if $SHOWME
 	then
 	    echo "+ $PMLOGCONF -c -q -h $hostname $configfile"
-	elif ! $PMLOGCONF -c -q -h $hostname "$configfile" >$tmp/diag 2>&1
+	elif ! $PMLOGCONF -c -q -h $hostname "$configfile" </dev/null >$tmp/diag 2>&1
 	then
 	    _warning "pmlogconf failed to generate \"$configfile\""
 	    cat $tmp/diag
@@ -389,9 +427,51 @@ _get_logfile()
 _get_primary_logger_pid()
 {
     pid=`cat "$PCP_RUN_DIR/pmlogger.pid" 2>/dev/null`
+    if [ -z "$pid" ]
+    then
+	# No PID file, try the pmcd.pmlogger.* info files where "primary"
+	# is a symlink to a <pid> file
+	#
+	pid=`ls -l $PCP_TMP_DIR/pmlogger/primary 2>/dev/null | sed -e 's;.*/\([0-9][0-9]*\)$;\1;'`
+    fi
     echo "$pid"
 }
 
+# wait for the local pmcd to get going for a primary pmlogger
+# (borrowed from qa/common.check)
+#
+# wait_for_pmcd [maxdelay]
+#
+_wait_for_pmcd()
+{
+    # 5 seconds default seems like a reasonable max time to get going
+    _can_wait=${1-5}
+    _limit=`expr $_can_wait \* 10`
+    _i=0
+    _dead=true
+    while [ $_i -lt $_limit ]
+    do
+	_sts=`pmprobe pmcd.numclients 2>/dev/null | $PCP_AWK_PROG '{print $2}'`
+	if [ "${_sts:-0}" -gt 0 ]
+	then
+	    # numval really > 0, we're done
+	    #
+	    _dead=false
+	    break
+	fi
+	pmsleep 0.1
+	_i=`expr $_i + 1`
+    done
+    if $_dead
+    then
+	date
+	echo "Arrgghhh ... pmcd at localhost failed to start after $_can_wait seconds"
+	echo "=== failing pmprobes ==="
+	pmprobe pmcd.numclients
+	status=1
+    fi
+}
+
 _check_archive()
 {
     if [ ! -e "$logfile" ]
@@ -528,7 +608,17 @@ _parse_control()
 	cd "$here"
 	line=`expr $line + 1`
 
-	$VERY_VERBOSE && echo "[$controlfile:$line] host=\"$host\" primary=\"$primary\" socks=\"$socks\" dir=\"$dir\" args=\"$args\""
+
+	if $VERY_VERBOSE 
+	then
+	    case "$host"
+	    in
+	    \#*|'')	# comment or empty
+			;;
+	    *)		echo "[$controlfile:$line] host=\"$host\" primary=\"$primary\" socks=\"$socks\" dir=\"$dir\" args=\"$args\""
+	    		;;
+	    esac
+	fi
 
 	case "$host"
 	in
@@ -596,6 +686,15 @@ s/^\([A-Za-z][A-Za-z0-9_]*\)=/export \1; \1=/p
 	    continue
 	fi
 
+	# if -s/--skip-primary on the command line, do not process
+	# a control file line for the primary pmlogger
+	#
+	if $SKIP_PRIMARY && [ $primary = y ]
+	then
+	    $VERY_VERBOSE && echo "Skip, -s/--skip-primary on command line"
+	    continue
+	fi
+
 	# substitute LOCALHOSTNAME marker in this config line
 	# (differently for directory and pcp -h HOST arguments)
 	#
@@ -607,7 +706,7 @@ s/^\([A-Za-z][A-Za-z0-9_]*\)=/export \1; \1=/p
 	then
 	    pflag=''
 	    [ $primary = y ] && pflag=' -P'
-	    echo "Check pmlogger$pflag -h $host ... in $dir ..."
+	    echo "Checking for: pmlogger$pflag -h $host ... in $dir ..."
 	fi
 
 	# check for directory duplicate entries
@@ -661,19 +760,25 @@ s/^\([A-Za-z][A-Za-z0-9_]*\)=/export \1; \1=/p
 	    delay=200	# tenths of a second
 	    while [ $delay -gt 0 ]
 	    do
-		if pmlock -v lock >$tmp/out 2>&1
+		if pmlock -v "$dir/lock" >$tmp/out 2>&1
 		then
-		    echo $dir/lock >$tmp/lock
+		    echo "$dir/lock" >$tmp/lock
+		    if $VERY_VERBOSE
+		    then
+			echo "Acquired lock:"
+			ls -l $dir/lock
+		    fi
 		    break
 		else
 		    [ -f $tmp/stamp ] || touch -t `pmdate -30M %Y%m%d%H%M` $tmp/stamp
-		    if [ -z "`find lock -newer $tmp/stamp -print 2>/dev/null`" ]
+		    find $tmp/stamp -newer "$dir/lock" -print 2>/dev/null >$tmp/tmp
+		    if [ -s $tmp/tmp ]
 		    then
-			if [ -f lock ]
+			if [ -f "$dir/lock" ]
 			then
 			    echo "$prog: Warning: removing lock file older than 30 minutes"
 			    LC_TIME=POSIX ls -l $dir/lock
-			    rm -f lock
+			    rm -f "$dir/lock"
 			else
 			    # there is a small timing window here where pmlock
 			    # might fail, but the lock file has been removed by
@@ -711,7 +816,7 @@ s/^\([A-Za-z][A-Za-z0-9_]*\)=/export \1; \1=/p
 			continue
 		    fi
 		fi
-		if [ -f lock ]
+		if [ -f "$dir/lock" ]
 		then
 		    echo "$prog: Warning: is another PCP cron job running concurrently?"
 		    LC_TIME=POSIX ls -l $dir/lock
@@ -724,7 +829,7 @@ s/^\([A-Za-z][A-Za-z0-9_]*\)=/export \1; \1=/p
 	fi
 
 	pid=''
-	if [ "X$primary" = Xy ]
+	if [ "$primary" = y ]
 	then
 	    if test -e "$PCP_TMP_DIR/pmlogger/primary"
 	    then
@@ -750,6 +855,14 @@ s/^\([A-Za-z][A-Za-z0-9_]*\)=/export \1; \1=/p
 		    $VERY_VERBOSE && echo "primary pmlogger process $pid not running"
 		    pid=''
 		fi
+	    else
+		if $VERY_VERBOSE
+		then
+		    echo "$PCP_TMP_DIR/pmlogger/primary: missing?"
+		    echo "Contents of $PCP_TMP_DIR/pmlogger"
+		    ls -l $PCP_TMP_DIR/pmlogger
+		    echo "--- end of ls output ---"
+		fi
 	    fi
 	else
 	    for log in $PCP_TMP_DIR/pmlogger/[0-9]*
@@ -762,7 +875,7 @@ s/^\([A-Za-z][A-Za-z0-9_]*\)=/export \1; \1=/p
 		    $PCP_ECHO_PROG $PCP_ECHO_N "... try $log host=$_host arch=$_arch: ""$PCP_ECHO_C"
 		fi
 		# throw away stderr in case $log has been removed by now
-		match=`sed -e '3s/\/[0-9][0-9][0-9][0-9][0-9.]*$//' $log 2>/dev/null | \
+		match=`sed -e '3s@/[^/]*$@@' $log 2>/dev/null | \
 		$PCP_AWK_PROG '
 BEGIN				{ m = 0 }
 NR == 3 && $0 == "'$dir'"	{ m = 2; next }
@@ -786,8 +899,6 @@ END				{ print m }'`
 
 	if [ -z "$pid" -a $START_PMLOGGER = true ]
 	then
-	    rm -f Latest
-
 	    if [ "X$primary" = Xy ]
 	    then
 		envs=`grep ^PMLOGGER "$PMLOGGERENVS" 2>/dev/null`
@@ -797,6 +908,17 @@ END				{ print m }'`
 		#
 		PM_LOG_PORT_DIR="$PCP_TMP_DIR/pmlogger"
 		rm -f "$PM_LOG_PORT_DIR/primary"
+		# We really expect the primary pmlogger to work, especially
+		# in the systemd world, so make sure pmcd is ready to accept
+		# connections.
+		#
+		_wait_for_pmcd
+		if [ "$status" = 1 ]
+		then
+		    $VERY_VERBOSE && echo "pmcd not running, skip primary pmlogger"
+		    _unlock "$dir"
+		    continue
+		fi
 	    else
 		args="-h $host $args"
 		envs=""
@@ -805,39 +927,13 @@ END				{ print m }'`
 
 	    # each new log started is named yyyymmdd.hh.mm
 	    #
-	    LOGNAME=`date "+%Y%m%d.%H.%M"`
+	    LOGNAME=%Y%m%d.%H.%M
 
-	    # handle duplicates/aliases (happens when pmlogger is restarted
-	    # within a minute and LOGNAME is the same)
+	    # We used to handle duplicates/aliases here (happens when
+	    # pmlogger is restarted within a minute and LOGNAME expands
+	    # to the same string ... this is now magically handled by
+	    # pmlogger, so do nothing.
 	    #
-	    suff=''
-	    for file in $LOGNAME.*
-	    do
-		[ "$file" = "$LOGNAME"'.*' ] && continue
-		# we have a clash! ... find a new -number suffix for the
-		# existing files ... we are going to keep $LOGNAME for the
-		# new pmlogger below
-		#
-		if [ -z "$suff" ]
-		then
-		    for xx in 0 1 2 3 4 5 6 7 8 9
-		    do
-			for yy in 0 1 2 3 4 5 6 7 8 9
-			do
-			    [ "`echo $LOGNAME-${xx}${yy}.*`" != "$LOGNAME-${xx}${yy}.*" ] && continue
-			    suff=${xx}${yy}
-			    break
-			done
-			[ ! -z "$suff" ] && break
-		    done
-		    if [ -z "$suff" ]
-		    then
-	    		_error "unable to break duplicate clash for archive basename $LOGNAME"
-		    fi
-		    $VERBOSE && echo "Duplicate archive basename ... rename $LOGNAME.* files to $LOGNAME-$suff.*"
-		fi
-		eval $MV -f $file `echo $file | sed -e "s/$LOGNAME/&-$suff/"`
-	    done
 
 	    configfile=`_get_configfile $args`
 	    if [ ! -z "$configfile" ]
@@ -886,55 +982,49 @@ END				{ print m }'`
 		eval $MV -f $logfile $logfile.prior
 	    fi
 
+	    # Notify service manager (if any) for the primary logger ONLY.
+	    # TODO if no primary, notify for the first logger.
+	    [ "$primary" = y ] && args="-N $args"
+
 	    args="$args -m pmlogger_check"
 	    if $SHOWME
 	    then
 		echo
 		echo "+ ${sock_me}$PMLOGGER $args $LOGNAME"
-		_unlock
+		_unlock "$dir"
 		continue
 	    else
 		$PCP_BINADM_DIR/pmpost "start pmlogger from $prog for host $host"
-		eval $envs '${sock_me}$PMLOGGER $args $LOGNAME >$tmp/out 2>&1 &'
-		pid=$!
+		# The pmlogger child will be re-parented to init (aka systemd)
+		pid=`eval $envs '${sock_me}$PMLOGGER $args $LOGNAME >$tmp/out 2>&1 & echo $!'`
 	    fi
 
 	    # wait for pmlogger to get started, and check on its health
 	    _check_logger $pid
 
-	    # the archive folio Latest is for the most recent archive in
-	    # this directory
-	    #
-	    if [ -f $LOGNAME.0 ] 
-	    then
-		$VERBOSE && echo "Latest folio created for $LOGNAME"
-		mkaf $LOGNAME.0 >Latest 2>/dev/null
-		chown $PCP_USER:$PCP_GROUP Latest >/dev/null 2>&1
-	    else
-		touch $tmp/err
-		logdir=`dirname $LOGNAME`
-		if $TERSE
-		then
-		    echo "$prog: Error: archive file `cd $logdir; $PWDCMND`/$LOGNAME.0 missing"
-		else
-		    echo "$prog: Error: archive file $LOGNAME.0 missing"
-		    echo "Directory (`cd $logdir; $PWDCMND`) contents:"
-		    LC_TIME=POSIX ls -la $logdir
-		fi
-	    fi
-
 	    # if SaveLogs exists in the same directory that the archive
 	    # is being created, save pmlogger log file there as well
 	    #
-	    dirname=`dirname $LOGNAME`
-	    if [ -d $dirname/SaveLogs ]
+	    if [ -d ./SaveLogs ]
 	    then
-		if [ ! -f $dirname/SaveLogs/$LOGNAME.log ]
+		# get archive basename, which is the expanded version
+		# of $LOGNAME, possibly with duplicate resolution ...
+		# the $PCP_TMP_DIR/pmlogger file has the answer
+		#
+		mylogname=`sed -n -e 3p $PCP_TMP_DIR/pmlogger/$pid 2>/dev/null \
+		           | sed -e 's;.*/;;'`
+		if [ -n "$mylogname" ]
 		then
-		    $LN $logfile $dirname/SaveLogs/$LOGNAME.log
+		    if [ ! -f ./SaveLogs/$mylogname.log ]
+		    then
+			$LN $logfile ./SaveLogs/$mylogname.log
+		    else
+			$VERBOSE && echo "Failed to link $logfile, SaveLogs/$mylogname.log already exists"
+		    fi
+		else
+		    $VERBOSE && echo "Failed to get archive basename from $PCP_TMP_DIR/pmlogger/$pid"
 		fi
 	    fi
-
 	elif [ ! -z "$pid" -a $STOP_PMLOGGER = true ]
 	then
 	    # Send pmlogger a SIGTERM, which is noted as a pending shutdown.
@@ -945,15 +1035,15 @@ END				{ print m }'`
 	    $PCP_ECHO_PROG $PCP_ECHO_N "$pid ""$PCP_ECHO_C" >> $tmp/pmloggers
 	fi
 
-	_unlock
+	_unlock "$dir"
     done
 }
 
-_parse_control $CONTROL
-append=`ls $CONTROLDIR 2>/dev/null | LC_COLLATE=POSIX sort`
-for controlfile in $append
+# parse and process the control file(s)
+append=`ls $CONTROLDIR 2>/dev/null | LC_COLLATE=POSIX sort | sed -e "s;^;$CONTROLDIR/;g"`
+for c in $CONTROL $append
 do
-    _parse_control $CONTROLDIR/$controlfile
+    _parse_control "$c"
 done
 
 # check all the SIGTERM'd loggers really died - if not, use a bigger hammer.
@@ -986,8 +1076,8 @@ then
 fi
 
 # Prior to exiting we compress existing logs, if any. See pmlogger_daily -K
-# Do not compress on a virgin install - it confuses systemd's Type=forking
-# exit status checks. There is nothing to compress anyway. See RHBZ#1721223.
+# Do not compress on a virgin install - there is nothing to compress anyway.
+# See RHBZ#1721223.
 [ -f "$PCP_LOG_DIR/pmlogger/pmlogger_daily.stamp" ] && _compress_now
 
 [ -f $tmp/err ] && status=1
