@@ -1,6 +1,6 @@
 #! /bin/sh
 #
-# Copyright (c) 2013-2016,2018 Red Hat.
+# Copyright (c) 2013-2016,2018,2020 Red Hat.
 # Copyright (c) 1995-2000,2003 Silicon Graphics, Inc.  All Rights Reserved.
 # 
 # This program is free software; you can redistribute it and/or modify it
@@ -26,24 +26,123 @@ unset PCP_STDERR
 
 # constant setup
 #
-tmp=`mktemp -d /tmp/pcp.XXXXXXXXX` || exit 1
+tmp=`mktemp -d "$PCP_TMPFILE_DIR/pmlogger_daily.XXXXXXXXX"` || exit 1
 status=0
 echo >$tmp/lock
 prog=`basename $0`
 PROGLOG=$PCP_LOG_DIR/pmlogger/$prog.log
+MYPROGLOG=$PROGLOG.$$
 USE_SYSLOG=true
 
 _cleanup()
 {
+    if [ -s "$MYPROGLOG" ]
+    then
+	rm -f "$PROGLOG"
+	mv "$MYPROGLOG" "$PROGLOG"
+    else
+	rm -f "$MYPROGLOG"
+    fi
     $USE_SYSLOG && [ $status -ne 0 ] && \
     $PCP_SYSLOG_PROG -p daemon.error "$prog failed - see $PROGLOG"
-    [ -s "$PROGLOG" ] || rm -f "$PROGLOG"
     lockfile=`cat $tmp/lock 2>/dev/null`
     rm -f "$lockfile" "$PCP_RUN_DIR/pmlogger_daily.pid"
     rm -rf $tmp
+    $VERY_VERBOSE && echo >&2 "End: `date '+%F %T.%N'`"
 }
 trap "_cleanup; exit \$status" 0 1 2 3 15
 
+# pretty format a find-filter timespec
+#
+_timespec()
+{
+    echo "$1" | sed -e 's/:/ /g' \
+    | while read _d _h _m
+    do
+	case "$_d:$_h:$_m"
+	in
+	    0::)
+		echo "0 days"
+		;;
+	    0::1)
+		echo "1 minute"
+		;;
+	    0::*)
+		echo "$_m minutes"
+		;;
+	    0:1:)
+		echo "1 hour"
+		;;
+	    0:*:)
+		echo "$_h hours"
+		;;
+	    0:1:1)
+		echo "1 hour and 1 minute"
+		;;
+	    0:1:*)
+		echo "1 hour and $_m minutes"
+		;;
+	    0:*:*)
+		echo "$_h hours and $_m minutes"
+		;;
+	    1::)
+		echo "1 day"
+		;;
+	    1::1)
+		echo "1 day and 1 minute"
+		;;
+	    1::*)
+		echo "1 day and $_m minutes"
+		;;
+	    1:1:)
+		echo "1 day and 1 hour"
+		;;
+	    1:*:)
+		echo "1 day and $_h hours"
+		;;
+	    1:1:1)
+		echo "1 day, 1 hour and 1 minute"
+		;;
+	    1:1:*)
+		echo "1 day, 1 hour and $_m minutes"
+		;;
+	    1:*:1)
+		echo "1 day, $_h hours and 1 minute"
+		;;
+	    1:*:*)
+		echo "1 day, $_h hours and $_m minutes"
+		;;
+	    *::)
+		echo "$_d days"
+		;;
+	    *::1)
+		echo "$_d days and 1 minute"
+		;;
+	    *::*)
+		echo "$_d days and $_m minutes"
+		;;
+	    *:1:)
+		echo "$_d days and 1 hour"
+		;;
+	    *:*:)
+		echo "$_d days and $_h hours"
+		;;
+	    *:1:1)
+		echo "$_d days, 1 hour and 1 minute"
+		;;
+	    *:1:*)
+		echo "$_d days, 1 hour and $_m minutes"
+		;;
+	    *:*:1)
+		echo "$_d days, $_h hours and 1 minute"
+		;;
+	    *:*:*)
+		echo "$_d days, $_h hours and $_m minutes"
+		;;
+	esac
+    done
+}
+
 if is_chkconfig_on pmlogger
 then
     PMLOGGER_CTL=on
@@ -90,10 +189,10 @@ else
 fi
 if [ -n "$PCP_COMPRESSAFTER" ]
 then
-    check=`echo "$PCP_COMPRESSAFTER" | sed -e 's/[0-9]//g'`
-    if [ ! -z "$check" -a X"$check" != Xforever -a X"$check" != Xnever ]
+    $PCP_BINADM_DIR/find-filter </dev/null >/dev/null 2>&1 mtime "+$PCP_COMPRESSAFTER"
+    if [ $? != 0 -a X"$PCP_COMPRESSAFTER" != Xforever -a X"$PCP_COMPRESSAFTER" != Xnever ]
     then
-	echo "Error: \$PCP_COMPRESSAFTER value ($PCP_COMPRESSAFTER) must be numeric, \"forever\" or \"never\""
+	echo "Error: \$PCP_COMPRESSAFTER value ($PCP_COMPRESSAFTER) must be number, time, \"forever\" or \"never\""
 	status=1
 	exit
     fi
@@ -140,8 +239,9 @@ echo > $tmp/usage
 cat >> $tmp/usage <<EOF
 Options:
   -c=FILE,--control=FILE  pmlogger control file
+  -E,--expunge            expunge metrics with metadata inconsistencies when merging archives
   -f,--force              force actions (intended for QA, not production)
-  -k=N,--discard=N        remove archives after N days
+  -k=TIME,--discard=TIME  remove archives after TIME (format DD[:HH[:MM]])
   -K                      compress, but no other changes
   -l=FILE,--logfile=FILE  send important diagnostic messages to FILE
   -m=ADDRs,--mail=ADDRs   send daily NOTICES entries to email addresses
@@ -154,7 +254,7 @@ Options:
   -s=SIZE,--rotate=SIZE   rotate NOTICES file after reaching SIZE bytes
   -t=WANT                 implies -VV, keep verbose output trace for WANT days
   -V,--verbose            verbose output (multiple times for very verbose)
-  -x=N,--compress-after=N  compress archive data files after N days
+  -x=TIME,--compress-after=TIME  compress archive data files after TIME (format DD[:HH[:MM]])
   -X=PROGRAM,--compressor=PROGRAM  use PROGRAM for archive data file compression
   -Y=REGEX,--regex=REGEX  egrep filter when compressing files ["$COMPRESSREGEX_DEFAULT"]
   --help
@@ -180,7 +280,9 @@ TRACE=0
 RFLAG=false
 REWRITEALL=false
 MFLAG=false
+EXPUNGE=""
 FORCE=false
+KILL=pmsignal
 
 ARGS=`pmgetopt --progname=$prog --config=$tmp/usage -- "$@"`
 [ $? != 0 ] && exit 1
@@ -194,14 +296,16 @@ do
 		CONTROLDIR="$2.d"
 		shift
 		;;
+	-E)	EXPUNGE="-E"
+		;;
 	-f)	FORCE=true
 		;;
 	-k)	CULLAFTER="$2"
 		shift
-		check=`echo "$CULLAFTER" | sed -e 's/[0-9]//g'`
-		if [ ! -z "$check" -a X"$check" != Xforever -a X"$check" != Xnever ]
+		$PCP_BINADM_DIR/find-filter </dev/null >/dev/null 2>&1 mtime "+$CULLAFTER"
+		if [ $? != 0 -a X"$CULLAFTER" != Xforever -a X"$CULLAFTER" != Xnever ]
 		then
-		    echo "Error: -k value ($CULLAFTER) must be numeric, \"forever\" or \"never\""
+		    echo "Error: -k value ($CULLAFTER) must be number, time, \"forever\" or \"never\""
 		    status=1
 		    exit
 		fi
@@ -214,8 +318,10 @@ do
 		fi
 		COMPRESSONLY=true
 		PROGLOG=$PCP_LOG_DIR/pmlogger/$prog-K.log
+		MYPROGLOG=$PROGLOG.$$
 		;;
 	-l)	PROGLOG="$2"
+		MYPROGLOG=$PROGLOG.$$
 		USE_SYSLOG=false
 		shift
 		;;
@@ -277,6 +383,7 @@ do
 		# $PCP_LOG_DIR/pmlogger/daily.<date>.trace
 		#
 		PROGLOG=$PCP_LOG_DIR/pmlogger/daily.`date "+%Y%m%d.%H.%M"`.trace
+		MYPROGLOG=$PROGLOG.$$
 		VERBOSE=true
 		VERY_VERBOSE=true
 		MYARGS="$MYARGS -V -V"
@@ -297,10 +404,10 @@ do
 		    COMPRESSAFTER_CMDLINE=""
 		    continue
 		fi
-		check=`echo "$COMPRESSAFTER_CMDLINE" | sed -e 's/[0-9]//g'`
-		if [ ! -z "$check" -a X"$check" != Xforever -a X"$check" != Xnever ]
+		$PCP_BINADM_DIR/find-filter </dev/null >/dev/null 2>&1 mtime "+$COMPRESSAFTER_CMDLINE"
+		if [ $? != 0 -a X"$COMPRESSAFTER_CMDLINE" != Xforever -a X"$COMPRESSAFTER_CMDLINE" != Xnever ]
 		then
-		    echo "Error: -x value ($COMPRESSAFTER_CMDLINE) must be numeric, \"forever\" or \"never\""
+		    echo "Error: -x value ($COMPRESSAFTER_CMDLINE) must be number, time, \"forever\" or \"never\""
 		    status=1
 		    exit
 		fi
@@ -417,13 +524,23 @@ else
     #
     # Exception ($SHOWME, above) is for -N where we want to see the output.
     #
-    touch "$PROGLOG"
-    chown $PCP_USER:$PCP_GROUP "$PROGLOG" >/dev/null 2>&1
-    exec 1>"$PROGLOG" 2>&1
+    touch "$MYPROGLOG"
+    chown $PCP_USER:$PCP_GROUP "$MYPROGLOG" >/dev/null 2>&1
+    exec 1>"$MYPROGLOG" 2>&1
+fi
+
+if $VERY_VERBOSE
+then
+    echo >&2 "Start: `date '+%F %T.%N'`"
+    if which pstree >/dev/null 2>&1
+    then
+	echo >&2 "Called from:"
+	pstree >&2 -spa $$
+    fi
 fi
 
 # if SaveLogs exists in the $PCP_LOG_DIR/pmlogger directory then save
-# $PROGLOG there as well with a unique name that contains the date and time
+# $MYPROGLOG there as well with a unique name that contains the date and time
 # when we're run ... skip if -N (showme)
 #
 if $SHOWME
@@ -432,15 +549,15 @@ then
 else
     if [ -d $PCP_LOG_DIR/pmlogger/SaveLogs ]
     then
-	now="`date '+%Y%m%d.%H.%M'`"
-	link=`echo $PROGLOG | sed -e "s/$prog/SaveLogs\/$prog.$now/"`
+	now="`date '+%Y%m%d.%H.%M.%S'`"
+	link=`echo $MYPROGLOG | sed -e "s/$prog/SaveLogs\/$prog.$now/"`
 	if [ ! -f "$link" ]
 	then
 	    if $SHOWME
 	    then
-		echo "+ ln $PROGLOG $link"
+		echo "+ ln $MYPROGLOG $link"
 	    else
-		ln $PROGLOG $link
+		ln $MYPROGLOG $link
 	    fi
 	fi
     fi
@@ -453,11 +570,6 @@ then
     exit
 fi
 
-# each new archive log started by pmnewlog or pmlogger_check is named
-# yyyymmdd.hh.mm
-#
-LOGNAME=`date "+%Y%m%d.%H.%M"`
-
 _error()
 {
     echo "$prog: [$filename:$line]"
@@ -491,19 +603,20 @@ _lock()
 	delay=200	# tenths of a second
 	while [ $delay -gt 0 ]
 	do
-	    if pmlock -v lock >>$tmp/out 2>&1
+	    if pmlock -v "$1/lock" >>$tmp/out 2>&1
 	    then
-		echo $1/lock >$tmp/lock
+		echo "$1/lock" >$tmp/lock
 		break
 	    else
 		[ -f $tmp/stamp ] || touch -t `pmdate -30M %Y%m%d%H%M` $tmp/stamp
-		if [ ! -z "`find lock -newer $tmp/stamp -print 2>/dev/null`" ]
+		find $tmp/stamp -newer "$1/lock" -print 2>/dev/null >$tmp/tmp
+		if [ -s $tmp/tmp ]
 		then
-		    if [ -f lock ]
+		    if [ -f "$1/lock" ]
 		    then
 			_warning "removing lock file older than 30 minutes"
-			LC_TIME=POSIX ls -l $1/lock
-			rm -f lock
+			LC_TIME=POSIX ls -l "$1/lock"
+			rm -f "$1/lock"
 		    else
 			# there is a small timing window here where pmlock
 			# might fail, but the lock file has been removed by
@@ -521,10 +634,10 @@ _lock()
 	then
 	    # failed to gain mutex lock
 	    #
-	    if [ -f lock ]
+	    if [ -f "$1/lock" ]
 	    then
 		_warning "is another PCP cron job running concurrently?"
-		LC_TIME=POSIX ls -l $1/lock
+		LC_TIME=POSIX ls -l "$1/lock"
 	    else
 		echo "$prog: `cat $tmp/out`"
 	    fi
@@ -538,7 +651,7 @@ _lock()
 
 _unlock()
 {
-    rm -f lock
+    rm -f "$1/lock"
     echo >$tmp/lock
 }
 
@@ -707,6 +820,9 @@ rm -f $tmp/err
 # if the directory containing the archive matches, then the name
 # of the file is the pid.
 #
+# The pid(s) (if any) appear on stdout, so be careful to send any
+# diagnostics to stderr.
+#
 _get_non_primary_logger_pid()
 {
     pid=''
@@ -717,27 +833,27 @@ _get_non_primary_logger_pid()
 	then
 	    _host=`sed -n 2p <$log`
 	    _arch=`sed -n 3p <$log`
-	    $PCP_ECHO_PROG $PCP_ECHO_N "... try $log host=$_host arch=$_arch: ""$PCP_ECHO_C"
+	    $PCP_ECHO_PROG >&2 $PCP_ECHO_N "... try $log host=$_host arch=$_arch: ""$PCP_ECHO_C"
 	fi
 	# throw away stderr in case $log has been removed by now
-	match=`sed -e '3s/\/[0-9][0-9][0-9][0-9][0-9.]*$//' $log 2>/dev/null | \
+	match=`sed -e '3s@/[^/]*$@@' $log 2>/dev/null | \
 	$PCP_AWK_PROG '
 BEGIN				{ m = 0 }
 NR == 3 && $0 == "'$dir'"	{ m = 2; next }
 END				{ print m }'`
-	$VERY_VERBOSE && $PCP_ECHO_PROG $PCP_ECHO_N "match=$match ""$PCP_ECHO_C"
+	$VERY_VERBOSE && $PCP_ECHO_PROG >&2 $PCP_ECHO_N "match=$match ""$PCP_ECHO_C"
 	if [ "$match" = 2 ]
 	then
 	    pid=`echo $log | sed -e 's,.*/,,'`
 	    if _get_pids_by_name pmlogger | grep "^$pid\$" >/dev/null
 	    then
-		$VERY_VERBOSE && echo "pmlogger process $pid identified, OK"
+		$VERY_VERBOSE && echo >&2 "pmlogger process $pid identified, OK"
 		break
 	    fi
-	    $VERY_VERBOSE && echo "pmlogger process $pid not running, skip"
+	    $VERY_VERBOSE && echo >&2 "pmlogger process $pid not running, skip"
 	    pid=''
 	else
-	    $VERY_VERBOSE && echo "different directory, skip"
+	    $VERY_VERBOSE && echo >&2 "different directory, skip"
 	fi
     done
     echo "$pid"
@@ -765,7 +881,7 @@ _parse_control()
 	$controlfile | \
     while read host primary socks dir args
     do
-	# start in one place for each iteration (beware relative paths)
+	# start in one place for each iteration (beware of relative paths)
 	cd "$here"
 	line=`expr $line + 1`
 
@@ -773,16 +889,22 @@ _parse_control()
 	then
 	    case "$host"
 	    in
+		\#!#*)	# stopped by pmlogctl ... we'll be checking this one
+			echo >&2 "[$filename:$line] host=\"$host\" primary=\"$primary\" socks=\"$socks\" dir=\"$dir\" args=\"$args\""
+			;;
 		\#*|'')	# comment or empty
 			;;
 		*)
-			echo "[$filename:$line] host=\"$host\" primary=\"$primary\" socks=\"$socks\" dir=\"$dir\" args=\"$args\""
+			echo >&2 "[$filename:$line] host=\"$host\" primary=\"$primary\" socks=\"$socks\" dir=\"$dir\" args=\"$args\""
 			;;
 	    esac
 	fi
 
 	case "$host"
 	in
+	    \#!#*)	# stopped by pmlogctl ... need to check this one
+		host=`echo "$host" | sed -e 's/^#!#//'`
+		;;
 	    \#*|'')	# comment or empty
 		continue
 		;;
@@ -831,10 +953,11 @@ s/^\([A-Za-z][A-Za-z0-9_]*\)=/export \1; \1=/p
 
 			'export PCP_COMPRESSAFTER;'*)
 			    old_value="$PCP_COMPRESSAFTER"
-			    check=`echo "$cmd" | sed -e 's/.*=//' -e 's/[0-9]//g' -e 's/  *$//'`
-			    if [ ! -z "$check" -a X"$check" != Xforever -a X"$check" != Xnever ]
+			    check=`echo "$cmd" | sed -e 's/.*=//' -e 's/  *$//'`
+			    $PCP_BINADM_DIR/find-filter </dev/null >/dev/null 2>&1 mtime "+$check"
+			    if [ $? != 0 -a -n "$check" -a X"$check" != Xforever -a X"$check" != Xnever ]
 			    then
-				_error "\$PCP_COMPRESSAFTER value ($check) must be numeric, \"forever\" or \"never\""
+				_error "\$PCP_COMPRESSAFTER value ($check) must be number, time, \"forever\" or \"never\""
 			    else
 				$SHOWME && echo "+ $cmd"
 				echo eval $cmd >>$tmp/cmd
@@ -911,11 +1034,23 @@ s/^\([A-Za-z][A-Za-z0-9_]*\)=/export \1; \1=/p
 	dir=`echo $dir | sed -e "s;LOCALHOSTNAME;$dirhostname;"`
 	[ $primary = y -o "x$host" = xLOCALHOSTNAME ] && host=local:
 
+	if $VERBOSE
+	then
+	    echo
+	    if $COMPRESSONLY
+	    then
+		echo "=== compressing PCP archives for host $host ==="
+	    else
+		echo "=== daily maintenance of PCP archives for host $host ==="
+	    fi
+	    echo
+	fi
+
 	if $VERY_VERBOSE
 	then
 	    pflag=''
 	    [ $primary = y ] && pflag=' -P'
-	    echo "Check pmlogger$pflag -h $host ... in $dir ..."
+	    echo >&2 "Check pmlogger$pflag -h $host ... in $dir ..."
 	fi
 
 	# make sure output directory hierarchy exists and $PCP_USER
@@ -938,18 +1073,6 @@ s/^\([A-Za-z][A-Za-z0-9_]*\)=/export \1; \1=/p
 	dir=`$PWDCMND`
 	$SHOWME && echo "+ cd $dir"
 
-	if $VERBOSE
-	then
-	    echo
-	    if $COMPRESSONLY
-	    then
-		echo "=== compressing PCP archives for host $host ==="
-	    else
-		echo "=== daily maintenance of PCP archives for host $host ==="
-	    fi
-	    echo
-	fi
-
 	if $SHOWME
 	then
 	    echo "+ get mutex lock"
@@ -964,7 +1087,7 @@ s/^\([A-Za-z][A-Za-z0-9_]*\)=/export \1; \1=/p
 	    fi
 	fi
 
-	# For archive rewiting (to make metadata consistent across
+	# For archive rewriting (to make metadata consistent across
 	# archives) find the rules as follows:
 	# - if pmlogrewrite exists (as a file, directory or symlink)
 	#   in the current archive directory use that
@@ -989,7 +1112,7 @@ s/^\([A-Za-z][A-Za-z0-9_]*\)=/export \1; \1=/p
 	    rewrite_args="$rewrite"
 	    if $VERBOSE
 	    then
-		echo "$prog: Info: pmlogrewrite all archives in $dir"
+		echo "Info: pmlogrewrite all archives in $dir"
 		rewrite_args="$rewrite_args -V"
 	    fi
 	    $VERY_VERBOSE && rewrite_args="$rewrite_args -V"
@@ -1013,26 +1136,37 @@ s/^\([A-Za-z][A-Za-z0-9_]*\)=/export \1; \1=/p
 	    then
 		_host=`sed -n 2p <"$PCP_TMP_DIR/pmlogger/primary"`
 		_arch=`sed -n 3p <"$PCP_TMP_DIR/pmlogger/primary"`
-		$VERY_VERBOSE && echo "... try $PCP_TMP_DIR/pmlogger/primary: host=$_host arch=$_arch"
+		$VERY_VERBOSE && echo >&2 "... try $PCP_TMP_DIR/pmlogger/primary: host=$_host arch=$_arch"
 		pid=`_get_primary_logger_pid`
 	    fi
 	    if [ -z "$pid" ]
 	    then
 		if $VERY_VERBOSE
 		then
-		    echo "primary pmlogger process PID not found"
-		    ls -l "$PCP_TMP_DIR/pmlogger"
-		    $PCP_PS_PROG $PCP_PS_ALL_FLAGS | egrep '[P]ID|[p]mlogger'
+		    echo >&2 "primary pmlogger process PID not found"
+		    ls >&2 -l "$PCP_TMP_DIR/pmlogger"
+		    $PCP_PS_PROG $PCP_PS_ALL_FLAGS | egrep >&2 '[P]ID|[p]mlogger'
 		fi
 	    elif _get_pids_by_name pmlogger | grep "^$pid\$" >/dev/null
 	    then
-		$VERY_VERBOSE && echo "primary pmlogger process $pid identified, OK"
+		$VERY_VERBOSE && echo >&2 "primary pmlogger process $pid identified, OK"
 	    else
-		$VERY_VERBOSE && echo "primary pmlogger process $pid not running"
+		$VERY_VERBOSE && echo >&2 "primary pmlogger process $pid not running"
 		pid=''
 	    fi
 	else
+	    # pid(s) on stdout, diagnostics on stderr
+	    #
 	    pid=`_get_non_primary_logger_pid`
+	    if $VERY_VERBOSE
+	    then
+		if [ -z "$pid" ]
+		then
+		    $VERY_VERBOSE && echo >&2 "No non-primary pmlogger process(es) found"
+		else
+		    $VERY_VERBOSE && echo >&2 "non-primary pmlogger process(es) $pid identified, OK"
+		fi
+	    fi
 	fi
 
 	if [ -z "$pid" ]
@@ -1043,60 +1177,16 @@ s/^\([A-Za-z][A-Za-z0-9_]*\)=/export \1; \1=/p
 	    else
 		_warning "no pmlogger instance running for host \"$host\""
 	    fi
-	    _warning "skipping pmnewlog because we don't know which pmlogger to stop"
+	    _warning "skipping log rotation because we don't know which pmlogger to signal"
 	elif ! $COMPRESSONLY
 	then
-	    # execute pmnewlog to "roll the archive logs"
+	    # send pmlogger a SIGUSR2 to "roll the archive logs"
 	    #
-	    [ X"$primary" != Xy ] && args="-p $pid $args"
-	    # else: -P is already the default
-	    [ X"$socks" = Xy ] && args="-s $args"
-	    args="$args -m pmlogger_daily"
-	    $SHOWME && echo "+ pmnewlog$MYARGS $args $LOGNAME"
-	    pid=''
-	    if pmnewlog$MYARGS $args $LOGNAME
+	    if $SHOWME
 	    then
-		:
+		echo "+ $KILL -s USR2 $pid"
 	    else
-		_error "problems executing pmnewlog for host \"$host\""
-	    fi
-	    # determine pid for (hopefully new) pmlogger instance for this
-	    # control line
-	    if [ X"$primary" = Xy ]
-	    then
-		if test -e "$PCP_TMP_DIR/pmlogger/primary"
-		then
-		    _host=`sed -n 2p <"$PCP_TMP_DIR/pmlogger/primary"`
-		    _arch=`sed -n 3p <"$PCP_TMP_DIR/pmlogger/primary"`
-		    $VERY_VERBOSE && echo "... try new $PCP_TMP_DIR/pmlogger/primary: host=$_host arch=$_arch"
-		    pid=`_get_primary_logger_pid`
-		fi
-		if [ -z "$pid" ]
-		then
-		    if $VERY_VERBOSE
-		    then
-			echo "new primary pmlogger process PID not found"
-			ls -l "$PCP_TMP_DIR/pmlogger"
-			$PCP_PS_PROG $PCP_PS_ALL_FLAGS | egrep '[P]ID|[p]mlogger'
-		    fi
-		elif _get_pids_by_name pmlogger | grep "^$pid\$" >/dev/null
-		then
-		    $VERY_VERBOSE && echo "new primary pmlogger process $pid identified, OK"
-		else
-		    $VERY_VERBOSE && echo "new primary pmlogger process $pid not running"
-		    pid=''
-		fi
-	    else
-		pid=`_get_non_primary_logger_pid`
-		if [ -z "$pid" ]
-		then
-		    if $VERY_VERBOSE
-		    then
-			echo "new non-primary pmlogger process PID not found"
-			ls -l "$PCP_TMP_DIR/pmlogger"
-			$PCP_PS_PROG $PCP_PS_ALL_FLAGS | egrep '[P]ID|[p]mlogger'
-		    fi
-		fi
+		$KILL -s USR2 "$pid"
 	    fi
 	fi
 
@@ -1111,27 +1201,57 @@ s/^\([A-Za-z][A-Za-z0-9_]*\)=/export \1; \1=/p
 	    #
 	    if [ X"$CULLAFTER" != Xforever -a X"$CULLAFTER" != Xnever ]
 	    then
+		# *BSD semantics for find(1) -mtime +N are "rounded up to
+		# the next full 24-hour period", compared to GNU/Linux
+		# semantics "any fractional part is ignored".  So, these are
+		# almost always off by one day in terms of the files selected.
+		# 
+		# For GNU/Linux reduce the number of days by one (opens up
+		# the window).
+		#
+		# The real filtering is done by find-filter, we just want to
+		# be sure that we include all of the candidate files in the
+		# find(1) part.
+		#
+		CULLDAYS=`echo $CULLAFTER | sed -e 's/:.*//'`
 		if [ "$PCP_PLATFORM" = freebsd -o "$PCP_PLATFORM" = netbsd -o "$PCP_PLATFORM" = openbsd ]
 		then
-		    # *BSD semantics for find(1) -mtime +N are "rounded up to
-		    # the next full 24-hour period", compared to GNU/Linux
-		    # semantics "any fractional part is ignored".  So, these are
-		    # almost always off by one day in terms of the files selected.
-		    # For consistency, try to match the GNU/Linux semantics by
-		    # using one MORE day.
-		    #
-		    mtime=`expr $CULLAFTER + 1`
+		    :
 		else
-		    mtime=$CULLAFTER
+		    # decrement, unless already zero
+		    #
+		    [ "$CULLDAYS" -gt 0 ] && CULLDAYS=`expr "$CULLDAYS" - 1`
 		fi
-		find . -type f -mtime +$mtime \
+		if [ "$CULLDAYS" -gt 0 ]
+		then
+		    find . -type f -mtime +$CULLDAYS
+		else
+		    find . -type f
+		fi \
+		| $PCP_BINADM_DIR/find-filter mtime +$CULLAFTER \
 		| _filter_filename \
 		| sort >$tmp/list
+		if [ -n "$pid" ]
+		then
+		    # pmlogger is running, make sure we don't cull the current
+		    # archive
+		    #
+		    if [ -f "$PCP_TMP_DIR/pmlogger/$pid" ]
+		    then
+			current_base=`sed -n -e '3s;.*/;;p' <"$PCP_TMP_DIR/pmlogger/$pid"`
+			if [ -n "$current_base" ]
+			then
+			    $VERY_VERBOSE && echo >&2 "[$filename:$line] skip cull for current $current_base archive"
+			    sed -e "/^$current_base/d" <$tmp/list >$tmp/tmp
+			    mv $tmp/tmp $tmp/list
+			fi
+		    fi
+		fi
 		if [ -s $tmp/list ]
 		then
 		    if $VERBOSE
 		    then
-			echo "Archive files older than $CULLAFTER days being removed ..."
+			echo "Archive files older than `_timespec $CULLAFTER` being removed ..."
 			fmt <$tmp/list | sed -e 's/^/    /'
 		    fi
 		    if $SHOWME
@@ -1141,7 +1261,7 @@ s/^\([A-Za-z][A-Za-z0-9_]*\)=/export \1; \1=/p
 			cat $tmp/list | xargs rm -f
 		    fi
 		else
-		    $VERY_VERBOSE && echo "$prog: Warning: no archive files found to cull"
+		    $VERY_VERBOSE && echo >&2 "Warning: no archive files found to cull"
 		fi
 	    fi
 
@@ -1210,10 +1330,15 @@ END	{ if (inlist != "") print lastdate,inlist }' >$tmp/list
 	    else
 		if [ ! -s $tmp/list ]
 		then
-		    if $VERBOSE
+		    if $SHOWME
 		    then
-			echo "$prog: Warning: no archives found to merge"
-			$VERY_VERBOSE && ls -l
+			:
+		    else
+			if $VERBOSE
+			then
+			    echo "Warning: no archives found to merge"
+			    $VERY_VERBOSE && ls >&2 -l
+			fi
 		    fi
 		else
 		    cat $tmp/list \
@@ -1224,7 +1349,7 @@ END	{ if (inlist != "") print lastdate,inlist }' >$tmp/list
 			    _skipping "output archive ($outfile) already exists"
 			    continue
 			else
-			    $VERY_VERBOSE && echo "Rewriting input archives using $rewrite"
+			    $VERY_VERBOSE && echo >&2 "Rewriting input archives using $rewrite"
 			    if $RFLAG
 			    then
 				:
@@ -1250,12 +1375,12 @@ END	{ if (inlist != "") print lastdate,inlist }' >$tmp/list
 			    then
 				for arch in $inlist
 				do
-				    echo "Input archive $arch ..."
+				    echo >&2 "Input archive $arch ..."
 				    if $SHOWME
 				    then
 					echo "+ pmdumplog -L $arch"
 				    else
-					pmdumplog -L $arch
+					pmdumplog >&2 -L $arch
 				    fi
 				done
 			    fi
@@ -1271,8 +1396,8 @@ END	{ if (inlist != "") print lastdate,inlist }' >$tmp/list
 				then
 				    if $VERY_VERBOSE
 				    then
-					echo "Renamed output archive $outfile ..."
-					pmdumplog -L $outfile
+					echo >&2 "Renamed output archive $outfile ..."
+					pmdumplog >&2 -L $outfile
 				    fi
 				else
 				    _error "problems executing pmlogmv for host \"$host\""
@@ -1282,13 +1407,13 @@ END	{ if (inlist != "") print lastdate,inlist }' >$tmp/list
 				#
 				if $SHOWME
 				then
-				    echo "+ pmlogger_merge$MYARGS -f $inlist $outfile"
-				elif pmlogger_merge$MYARGS -f $inlist $outfile
+				    echo "+ pmlogger_merge$MYARGS $EXPUNGE -f $inlist $outfile"
+				elif pmlogger_merge$MYARGS $EXPUNGE -f $inlist $outfile
 				then
 				    if $VERY_VERBOSE
 				    then
-					echo "Merged output archive $outfile ..."
-					pmdumplog -L $outfile
+					echo >&2 "Merged output archive $outfile ..."
+					pmdumplog >&2 -L $outfile
 				    fi
 				else
 				    _error "problems executing pmlogger_merge for host \"$host\""
@@ -1306,7 +1431,7 @@ END	{ if (inlist != "") print lastdate,inlist }' >$tmp/list
 	COMPRESSAFTER="$PCP_COMPRESSAFTER"
 	[ -z "$COMPRESSAFTER" ] && COMPRESSAFTER="$COMPRESSAFTER_CMDLINE"
 	[ -z "$COMPRESSAFTER" ] && COMPRESSAFTER="$COMPRESSAFTER_DEFAULT"
-	$VERY_VERBOSE && echo "$prog: COMPRESSAFTER=$COMPRESSAFTER"
+	$VERY_VERBOSE && echo >&2 "COMPRESSAFTER=$COMPRESSAFTER"
 	if [ -n "$COMPRESSAFTER" -a X"$COMPRESSAFTER" != Xforever -a X"$COMPRESSAFTER" != Xnever ]
 	then
 	    # may have some compression to do ...
@@ -1396,7 +1521,7 @@ p
 			# either the current index or the current metadata
 			# files
 			#
-			$VERY_VERBOSE && echo "[$filename:$line] skip current vol $current_base.$current_vol"
+			$VERY_VERBOSE && echo >&2 "[$filename:$line] skip compress for current $current_base.$current_vol volume"
 			rm -f $tmp/out
 			touch $tmp/out
 			# need to handle both the year 2000 and the old name
@@ -1450,7 +1575,7 @@ p
 			    cat $tmp/list | xargs $COMPRESS
 			fi
 		    else
-			$VERY_VERBOSE && echo "$prog: Warning: no archive files found to compress"
+			$VERY_VERBOSE && echo >&2 "Warning: no archive files found to compress"
 		    fi
 		elif [ -z "$COMPRESSAFTER" -o X"$COMPRESSAFTER" = Xforever -o X"$COMPRESSAFTER" = Xnever ]
 		then
@@ -1493,11 +1618,11 @@ p
 		    cat $tmp/list | xargs rm -f
 		fi
 	    else
-		$VERY_VERBOSE && echo "$prog: Warning: no trace files found to cull"
+		$VERY_VERBOSE && echo >&2 "Warning: no trace files found to cull"
 	    fi
 	fi
 
-	_unlock
+	_unlock "$dir"
     done
 }
 
@@ -1509,7 +1634,7 @@ p
 if [ -f $PCP_LOG_DIR/pmlogger/.NeedRewrite ]
 then
     REWRITEALL=true
-    $VERBOSE && echo "$prog: Info: found .NeedRewrite => rewrite all archives"
+    $VERBOSE && echo "Info: found .NeedRewrite => rewrite all archives"
 fi
 
 _parse_control $CONTROL
