@@ -1,6 +1,6 @@
 #! /bin/sh
 #
-# Copyright (c) 2014 Red Hat.
+# Copyright (c) 2014,2020 Red Hat.
 # Copyright (c) 1995,2003 Silicon Graphics, Inc.  All Rights Reserved.
 # 
 # This program is free software; you can redistribute it and/or modify it
@@ -25,13 +25,15 @@
 
 
 prog=`basename $0`
-tmp=`mktemp -d /tmp/pcp.XXXXXXXXX` || exit 1
+tmp=`mktemp -d "$PCP_TMPFILE_DIR/pmlogger_merge.XXXXXXXXX"` || exit 1
+tmpmerge=`mktemp -d $PCP_TMPFILE_DIR/pmlogger_merge.XXXXXXXXX` || exit 1
 status=0
-trap "rm -rf $tmp; exit \$status" 0 1 2 3 15
+trap "rm -rf $tmp $tmpmerge; exit \$status" 0 1 2 3 15
 
 force=false
 VERBOSE=false
 SHOWME=false
+EXPUNGE=""
 RM=rm
 
 _abandon()
@@ -47,6 +49,13 @@ _warning()
     force=false
 }
 
+_usage()
+{
+    pmgetopt --usage --progname=$prog --config=$tmp/usage
+    status=1
+    exit
+}
+
 cat > $tmp/usage << EOF
 # Usage: [options] [input-basename ... output-name]
 
@@ -54,6 +63,7 @@ Options:
   -f, --force    remove input files after creating output files
   -N, --showme   perform a dry run, showing what would be done
   -V, --verbose  increase diagnostic verbosity
+  -E, --expunge  expunge metrics with metadata mismatches between archives
   --help
 EOF
 
@@ -77,12 +87,14 @@ do
 	-V)	VERBOSE=true
 		;;
 
+	-E)	EXPUNGE="-x" # for pmlogextract -x
+		;;
+
 	--)	shift
 		break
 		;;
 
-	-\?)	pmgetopt --usage --progname=$prog --config=$tmp/usage
-		_abandon
+	-\?)	_usage
 		;;
     esac
     shift
@@ -102,9 +114,7 @@ then
     done
     output="$1"
 else
-    pmgetopt --usage --progname=$prog --config=$tmp/usage
-    status=1
-    exit
+    _usage
 fi
 
 fail=false
@@ -229,8 +239,8 @@ else
 	    # output = 108 file descriptors which should be well below any
 	    # shell-imposed or system-imposed limits
 	    #
-	    $VERBOSE && echo "		-> partial merge to $tmp/$part"
-	    cmd="pmlogextract $list $tmp/$part"
+	    $VERBOSE && echo "		-> partial merge to $tmpmerge/$part"
+	    cmd="pmlogextract $EXPUNGE $list $tmpmerge/$part"
 	    if $SHOWME
 	    then
 		echo "+ $cmd"
@@ -239,13 +249,13 @@ else
 		then
 		    :
 		else
-		    $VERBOSE || echo "		-> partial merge to $tmp/$part"
+		    $VERBOSE || echo "		-> partial merge to $tmpmerge/$part"
 		    echo "$prog: Directory: `pwd`"
-		    echo "$prog: Failed: pmlogextract $list $tmp/$part"
+		    echo "$prog: Failed: pmlogextract $list $tmpmerge/$part"
 		    _warning
 		fi
 	    fi
-	    list=$tmp/$part
+	    list=$tmpmerge/$part
 	    part=`expr $part + 1`
 	    i=0
 	fi
@@ -261,7 +271,7 @@ else
 	i=`expr $i + 1`
     done
 
-    cmd="pmlogextract $list $output"
+    cmd="pmlogextract $EXPUNGE $list $output"
     if $SHOWME
     then
 	echo "+ $cmd"
