@@ -79,7 +79,7 @@ get_pid_from_symlink(const char *linkfile, pid_t *pidp)
     return -1;
 }
 
-static void
+void
 cleanup(void)
 {
     /*
@@ -172,12 +172,11 @@ sigpipe_handler(int sig)
 }
 
 static void
-sigusr1_handler(int sig)
+sigusr2_handler(int sig)
 {
-    /*
-     * no-op now that all archive write I/O is unbuffered
-     */
-    __pmSetSignalHandler(SIGUSR1, sigusr1_handler);
+    __pmSetSignalHandler(SIGUSR2, sigusr2_handler);
+    log_switch_flag = 1;
+    sig_code = sig; /* triggers break from main loop so we can re-exec */
 }
 #endif
 
@@ -210,8 +209,8 @@ static sig_map_t	sig_handler[] = {
 #endif
     { SIGTERM,	sigterm_handler },	/* Exit   Terminated */
 #ifndef IS_MINGW
-    { SIGUSR1,	sigusr1_handler },	/* NOP    User Signal 1 - [was fflush(3)] */
-    { SIGUSR2,	sigexit_handler },	/* Exit   User Signal 2 */
+    { SIGUSR1,	sigterm_handler },	/* Exit User Signal 1 */
+    { SIGUSR2,	sigusr2_handler },	/* reexec User Signal 2 */
     { SIGCHLD,	SIG_IGN },		/* NOP    Child stopped or terminated */
 #ifdef SIGPWR
     { SIGPWR,	SIG_DFL },		/* Ignore Power Fail/Restart */
@@ -470,16 +469,16 @@ GetPorts(char *file)
 
 	/* then the full pathname to the archive base */
 	/* THREADSAFE - no locks acquired in __pmNativePath() */
-	archBase = __pmNativePath(archBase);
-	if (__pmAbsolutePath(archBase))
-	    fprintf(mapstream, "%s\n", archBase);
+	archName = __pmNativePath(archName);
+	if (__pmAbsolutePath(archName))
+	    fprintf(mapstream, "%s\n", archName);
 	else {
 	    char		path[MAXPATHLEN];
 
 	    if (getcwd(path, MAXPATHLEN) == NULL)
 		fprintf(mapstream, "\n");
 	    else
-		fprintf(mapstream, "%s%c%s\n", path, pmPathSeparator(), archBase);
+		fprintf(mapstream, "%s%c%s\n", path, pmPathSeparator(), archName);
 	}
 
 	/* and finally, the annotation from -m or -x */
@@ -648,9 +647,9 @@ init_ports(void)
 	}
 
 	/*
-	 * If the symlink still exists then there really is another primary logger running
+	 * If we have not reexec'd and the symlink still exists, then there really is another primary logger running
 	 */
-	if (access(linkfile, F_OK) == 0) {
+	if (!pmlogger_reexec && access(linkfile, F_OK) == 0) {
 	    /* configuration error - only one primary pmlogger should be configured */
 	    if (pid == -1)
 		fprintf(stderr, "%s: ERROR: there is already a primary pmlogger running, pid <unknown> linkfile=%s\n",
