@@ -16,17 +16,23 @@
 #include "libpcp.h"
 #include "pmproxy.h"
 #include "pmwebapi.h"
+#include <strings.h>
+#include <sys/stat.h>
 
 #define MAXPENDING	128	/* maximum number of pending connections */
 #define STRINGIFY(s)    #s
 #define TO_STRING(s)    STRINGIFY(s)
 
+#define RUN_DAEMON	1		/* default */
+#define RUN_FOREGROUND	2		/* -f */
+#define RUN_SYSTEMD	3		/* -F */
+
 static void	*info;			/* opaque server information */
 static pmproxy	*server;		/* proxy server implementation */
 struct dict	*config;		/* configuration file settings */
 
 static char	*logfile = "pmproxy.log";	/* log file name */
-static int	run_daemon = 1;		/* run as a daemon, see -f */
+static int	run_mode = RUN_DAEMON;	/* style of execution, see -f and -F */
 static char	*fatalfile = "/dev/tty";/* fatal messages at startup go here */
 static char	*username;
 static char	*certdb;		/* certificate DB path (NSS) */
@@ -67,6 +73,7 @@ static pmLongOptions longopts[] = {
     { "", 0, 'A', 0, "disable service advertisement" },
     { "deprecated", 0, 'd', 0, "backward-compatibility mode; no REST APIs" },
     { "foreground", 0, 'f', 0, "run in the foreground" },
+    { "systemd", 0, 'F', 0, "run in systemd mode" },
     { "timeseries", 0, 't', 0, "automatic, scalable timeseries; REST APIs" },
     { "username", 1, 'U', "USER", "in daemon mode, run as named user [default pcp]" },
     PMAPI_OPTIONS_HEADER("Configuration options"),
@@ -88,7 +95,7 @@ static pmLongOptions longopts[] = {
 };
 
 static pmOptions opts = {
-    .short_options = "Ac:C:dD:fh:i:l:L:M:p:P:r:s:tU:x:?",
+    .short_options = "Ac:C:dD:Ffh:i:l:L:M:p:P:r:s:tU:x:?",
     .long_options = longopts,
 };
 
@@ -132,8 +139,24 @@ ParseOptions(int argc, char *argv[], int *nports, int *maxpending)
 	    }
 	    break;
 
+	case 'F':	/* foreground but do pidfile and uid work */
+	    if (run_mode != RUN_DAEMON) {
+		pmprintf("%s: at most one of -f and -F allowed\n",
+			pmGetProgname());
+		opts.errors++;
+	    }
+	    else
+		run_mode = RUN_SYSTEMD;
+	    break;
+
 	case 'f':	/* foreground, i.e. do _not_ run as a daemon */
-	    run_daemon = 0;
+	    if (run_mode != RUN_DAEMON) {
+		pmprintf("%s: at most one of -f and -F allowed\n",
+			pmGetProgname());
+		opts.errors++;
+	    }
+	    else
+		run_mode = RUN_FOREGROUND;
 	    break;
 
 	case 'h':	/* Redis host name */
@@ -307,6 +330,7 @@ main(int argc, char *argv[])
     int		env_warn = 0;
     int		timeseries;
     char	*envstr;
+    pid_t	mainpid;
 
     umask(022);
     pmGetUsername(&username);
@@ -328,7 +352,42 @@ main(int argc, char *argv[])
     }
     timeseries = ParseOptions(argc, argv, &nport, &maxpending);
 
-    pmOpenLog(pmGetProgname(), logfile, stderr, &sts);
+    if (pmDebugOptions.appl1) {
+	/*
+	 * -Dappl1 is desperate logging mode ... insert .<pid> into
+	 * the logfile name, just before the last ., so pmproxy.log
+	 * will become pmproxy.<pid>.log
+	 */
+	char	newlogfile[MAXPATHLEN];
+	char	pbuf[11];	/* enough for a 32-bit pid */
+	char	*pend = NULL;
+
+	snprintf(pbuf, sizeof(pbuf), ".%" FMT_PID, (pid_t)getpid());
+	pend = rindex(logfile, '.');
+	if (pend == NULL) {
+	    /* no '.', so append .<pid> */
+	    strncpy(newlogfile, logfile, MAXPATHLEN-1);
+	    strcat(newlogfile, pbuf);
+	}
+	else {
+	    /* stitch name together ... <pre>.<post> -> <pre>.<pid>.<post> */
+	    char	*q = newlogfile;
+	    char	*r = pbuf;
+	    char	*p;
+
+	    for (p = logfile; p < pend; ) 	/* <pre> */
+		*q++ = *p++;
+	    while (*r)				/* .<pid> */
+		*q++ = *r++;
+	    while (*p)				/* .<post> */
+		*q++ = *p++;
+	    *q = '\0';
+	}
+	pmOpenLog(pmGetProgname(), newlogfile, stderr, &sts);
+    }
+    else
+	pmOpenLog(pmGetProgname(), logfile, stderr, &sts);
+
     /* close old stdout, and force stdout into same stream as stderr */
     fflush(stdout);
     close(fileno(stdout));
@@ -347,8 +406,11 @@ main(int argc, char *argv[])
     /* Advertise the service on the network if that is supported */
     __pmServerSetServiceSpec(PM_SERVER_PROXY_SPEC);
 
-    if (run_daemon)
+    if (run_mode == RUN_DAEMON) {
+	/* daemonize - fork and parent exits, setsid */
 	__pmServerStart(argc, argv, 1);
+    }
+    mainpid = getpid();
 
     /* Open non-blocking request ports for client connections */
     if ((info = server->openports(sockpath, sizeof(sockpath), maxpending)) == NULL)
@@ -364,7 +426,9 @@ main(int argc, char *argv[])
 	fprintf(stderr, "%s: maxpending=%d from PMPROXY_MAXPENDING=%s in environment\n",
 			"Warning", maxpending, getenv("PMPROXY_MAXPENDING"));
 
-    if (run_daemon) {
+    if (run_mode == RUN_DAEMON || run_mode == RUN_SYSTEMD) {
+	/* notify service manager, if any, we are ready */
+	__pmServerNotifyServiceManagerReady(mainpid);
 	if (__pmServerCreatePIDFile(PM_SERVER_PROXY_SPEC, PM_FATAL_ERR) < 0)
 	    DontStart();
 	if (pmSetProcessIdentity(username) < 0)
@@ -375,7 +439,7 @@ main(int argc, char *argv[])
         __pmSecureServerCertificateSetup(certdb, dbpassfile, cert_nickname) < 0)
 	DontStart();
 
-    fprintf(stderr, "pmproxy: PID = %" FMT_PID, (pid_t)getpid());
+    fprintf(stderr, "pmproxy: PID = %" FMT_PID, mainpid);
     fprintf(stderr, ", PDU version = %u", PDU_VERSION);
 #ifdef HAVE_GETUID
     fprintf(stderr, ", user = %s (%d)\n", username, getuid());
@@ -385,6 +449,9 @@ main(int argc, char *argv[])
 
     /* Loop processing client connections and server responses */
     server->loop(info);
+
+    /* inform service manager and shutdown cleanly */
+    __pmServerNotifyServiceManagerStopping(mainpid);
     Shutdown();
     exit(0);
 }
